#!/usr/bin/python3 -u

#libs
import argparse
import os
import signal
import sys
from PyQt5.QtWidgets import QApplication
from PyQt5.QtGui import QIcon, QFontDatabase
from PyQt5.QtCore import QLocale, QTranslator, QTimer

#local imports
from gui_signaler import Signaler
from daemon_manager import DaemonManager
from gui_tools import (
    ArgParser, CommandLineArgs, initGuiTools, default_session_root,
    ErrDaemon, _translate, getCodeRoot)
from gui_server_thread import GUIServerThread
from gui_session import SignaledSession
import nsm_client
import ray

#import UIs
import ui_raysession
import ui_client_slot

#import Qt resources
import resources_rc


def signalHandler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        if session._daemon_manager.launched_before:
            if (CommandLineArgs.under_nsm
                    and session.server_status != ray.ServerStatus.OFF):
                session._main_win.terminate_request = True
                
                server = GUIServerThread.instance()
                if server:
                    server.abortSession()
            else:
                session._daemon_manager.stop()
            return
        
        session._main_win.terminate_request = True
        session._daemon_manager.stop()

if __name__ == '__main__':
    #set Qt Application
    app = QApplication(sys.argv)
    app.setApplicationName("RaySession")
    app.setApplicationVersion(ray.VERSION)
    app.setOrganizationName("RaySession")
    app.setWindowIcon(QIcon(':/scalable/raysession.svg'))
    app.setQuitOnLastWindowClosed(False)
    
    ### Translation process
    locale = QLocale.system().name()
    appTranslator = QTranslator()
    if appTranslator.load("%s/locale/raysession_%s" % (getCodeRoot(), locale)):
        app.installTranslator(appTranslator)
    
    QFontDatabase.addApplicationFont(":/fonts/Ubuntu-R.ttf")
    QFontDatabase.addApplicationFont(":fonts/Ubuntu-C.ttf")
    
    initGuiTools()
    
    #Add raysession/src/bin to $PATH to can use raysession after make, whitout install
    ray.addSelfBinToPath()
    
    #get arguments
    parser = ArgParser()
    
    #connect signals
    signal.signal(signal.SIGINT , signalHandler)
    signal.signal(signal.SIGTERM, signalHandler)
    
    #needed for signals SIGINT, SIGTERM
    timer = QTimer()
    timer.start(200)
    timer.timeout.connect(lambda: None)
    
    #build session
    server = GUIServerThread()
    session = SignaledSession()
        
    app.exec()
    
    server.stop()
    session.quit()
    
    del session
    del app



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

# JACK ctypes definitions for usage in python applications
# Copyright (C) 2010-2013 Filipe Coelho <falktx@falktx.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# For a full copy of the GNU General Public License see the COPYING file

# ------------------------------------------------------------------------------------------------------------
# Imports (Global)

from ctypes import (
    cdll, c_int, c_uint8, c_uint32, c_uint64, c_int32, Structure, 
    c_float, c_size_t, POINTER, c_double, c_float, ARRAY, 
    c_char_p, CFUNCTYPE, c_void_p, c_ulong, pointer)
from sys import platform

# ------------------------------------------------------------------------------------------------------------
# Load JACK shared library

try:
    if platform == "darwin":
        jacklib = cdll.LoadLibrary("libjack.dylib")
    elif platform in ("win32", "win64", "cygwin"):
        jacklib = cdll.LoadLibrary("libjack.dll")
    else:
        jacklib = cdll.LoadLibrary("libjack.so.0")
except:
    jacklib = None
    raise ImportError("JACK is not available in this system")

# ------------------------------------------------------------------------------------------------------------
# JACK2 test

try:
    if jacklib.jack_get_version_string:
        JACK2 = True
    else:
        JACK2 = False
except:
    JACK2 = False

# ------------------------------------------------------------------------------------------------------------
# Pre-Types

c_enum  = c_int
c_uchar = c_uint8

class _jack_port(Structure):
    _fields_ = []

class _jack_client(Structure):
    _fields_ = []

# ------------------------------------------------------------------------------------------------------------
# Defines

JACK_MAX_FRAMES      = 4294967295
JACK_LOAD_INIT_LIMIT = 1024
JACK_DEFAULT_AUDIO_TYPE = "32 bit float mono audio"
JACK_DEFAULT_MIDI_TYPE  = "8 bit raw midi"

# ------------------------------------------------------------------------------------------------------------
# Types

jack_nframes_t   = c_uint32
jack_port_id_t   = c_uint32
jack_time_t      = c_uint64
jack_unique_t    = c_uint64
jack_midi_data_t = c_uchar
jack_default_audio_sample_t = c_float

jack_options_t = c_enum # JackOptions
jack_status_t  = c_enum # JackStatus
jack_transport_state_t    = c_enum # JackTransportState
jack_position_bits_t      = c_enum # JackPositionBits
jack_session_event_type_t = c_enum # JackSessionEventType
jack_session_flags_t      = c_enum # JackSessionFlags
jack_custom_change_t      = c_enum # JackCustomChange
jack_latency_callback_mode_t = c_enum # JackLatencyCallbackMode

jack_port_t   = _jack_port
jack_client_t = _jack_client
jack_port_type_id_t = c_uint32 # JACK2 only

# enum JackOptions
JackNullOption    = 0x00
JackNoStartServer = 0x01
JackUseExactName  = 0x02
JackServerName    = 0x04
JackLoadName      = 0x08
JackLoadInit      = 0x10
JackSessionID     = 0x20
JackOpenOptions   = JackSessionID|JackServerName|JackNoStartServer|JackUseExactName
JackLoadOptions   = JackLoadInit|JackLoadName|JackUseExactName

# enum JackStatus
JackFailure       = 0x01
JackInvalidOption = 0x02
JackNameNotUnique = 0x04
JackServerStarted = 0x08
JackServerFailed  = 0x10
JackServerError   = 0x20
JackNoSuchClient  = 0x40
JackLoadFailure   = 0x80
JackInitFailure   = 0x100
JackShmFailure    = 0x200
JackVersionError  = 0x400
JackBackendError  = 0x800
JackClientZombie  = 0x1000

# enum JackLatencyCallbackMode
JackCaptureLatency  = 0
JackPlaybackLatency = 1

# enum JackPortFlags
JackPortIsInput    = 0x1
JackPortIsOutput   = 0x2
JackPortIsPhysical = 0x4
JackPortCanMonitor = 0x8
JackPortIsTerminal = 0x10
JackPortIsControlVoltage = 0x100

# enum JackTransportState
JackTransportStopped     = 0
JackTransportRolling     = 1
JackTransportLooping     = 2
JackTransportStarting    = 3
JackTransportNetStarting = 4 # JACK2 only

# enum JackPositionBits
JackPositionBBT      = 0x10
JackPositionTimecode = 0x20
JackBBTFrameOffset   = 0x40
JackAudioVideoRatio  = 0x80
JackVideoFrameOffset = 0x100
JACK_POSITION_MASK   = JackPositionBBT|JackPositionTimecode|JackBBTFrameOffset|JackAudioVideoRatio|JackVideoFrameOffset

# enum JackSessionEventType
JackSessionSave         = 1
JackSessionSaveAndQuit  = 2
JackSessionSaveTemplate = 3

# enum JackSessionFlags
JackSessionSaveError    = 0x01
JackSessionNeedTerminal = 0x02

# enum JackCustomChange
JackCustomRemoved  = 0
JackCustomAdded    = 1
JackCustomReplaced = 2

# ------------------------------------------------------------------------------------------------------------
# Structs

class jack_midi_event_t(Structure):
    _fields_ = [
        ("time", jack_nframes_t),
        ("size", c_size_t),
        ("buffer", POINTER(jack_midi_data_t))
    ]

class jack_latency_range_t(Structure):
    _fields_ = [
        ("min", jack_nframes_t),
        ("max", jack_nframes_t)
    ]

class jack_position_t(Structure):
    _fields_ = [
        ("unique_1", jack_unique_t),
        ("usecs", jack_time_t),
        ("frame_rate", jack_nframes_t),
        ("frame", jack_nframes_t),
        ("valid", jack_position_bits_t),
        ("bar", c_int32),
        ("beat", c_int32),
        ("tick", c_int32),
        ("bar_start_tick", c_double),
        ("beats_per_bar", c_float),
        ("beat_type", c_float),
        ("ticks_per_beat", c_double),
        ("beats_per_minute", c_double),
        ("frame_time", c_double),
        ("next_time", c_double),
        ("bbt_offset", jack_nframes_t),
        ("audio_frames_per_video_frame", c_float),
        ("video_offset", jack_nframes_t),
        ("padding", ARRAY(c_int32, 7)),
        ("unique_2", jack_unique_t)
    ]

class jack_session_event_t(Structure):
    _fields_ = [
        ("type", jack_session_event_type_t),
        ("session_dir", c_char_p),
        ("client_uuid", c_char_p),
        ("command_line", c_char_p),
        ("flags", jack_session_flags_t),
        ("future", c_uint32)
    ]

class jack_session_command_t(Structure):
    _fields_ = [
        ("uuid", c_char_p),
        ("client_name", c_char_p),
        ("command", c_char_p),
        ("flags", jack_session_flags_t)
    ]

# ------------------------------------------------------------------------------------------------------------
# Callbacks

JackLatencyCallback = CFUNCTYPE(None, jack_latency_callback_mode_t, c_void_p)
JackProcessCallback = CFUNCTYPE(c_int, jack_nframes_t, c_void_p)
JackThreadCallback  = CFUNCTYPE(c_void_p, c_void_p)
JackThreadInitCallback = CFUNCTYPE(None, c_void_p)
JackGraphOrderCallback = CFUNCTYPE(c_int, c_void_p)
JackXRunCallback       = CFUNCTYPE(c_int, c_void_p)
JackBufferSizeCallback = CFUNCTYPE(c_int, jack_nframes_t, c_void_p)
JackSampleRateCallback = CFUNCTYPE(c_int, jack_nframes_t, c_void_p)
JackPortRegistrationCallback   = CFUNCTYPE(None, jack_port_id_t, c_int, c_void_p)
JackClientRegistrationCallback = CFUNCTYPE(None, c_char_p, c_int, c_void_p)
JackClientRenameCallback = CFUNCTYPE(c_int, c_char_p, c_char_p, c_void_p) # JACK2 only
JackPortConnectCallback  = CFUNCTYPE(None, jack_port_id_t, jack_port_id_t, c_int, c_void_p)
JackPortRenameCallback   = CFUNCTYPE(c_int, jack_port_id_t, c_char_p, c_char_p, c_void_p) # JACK2 only
JackFreewheelCallback    = CFUNCTYPE(None, c_int, c_void_p)
JackShutdownCallback     = CFUNCTYPE(None, c_void_p)
JackInfoShutdownCallback = CFUNCTYPE(None, jack_status_t, c_char_p, c_void_p)
JackSyncCallback     = CFUNCTYPE(c_int, jack_transport_state_t, POINTER(jack_position_t), c_void_p)
JackTimebaseCallback = CFUNCTYPE(None, jack_transport_state_t, jack_nframes_t, POINTER(jack_position_t), c_int, c_void_p)
JackSessionCallback  = CFUNCTYPE(None, POINTER(jack_session_event_t), c_void_p)
JackCustomDataAppearanceCallback = CFUNCTYPE(None, c_char_p, c_char_p, jack_custom_change_t, c_void_p)

# ------------------------------------------------------------------------------------------------------------
# Functions

try:
    jacklib.jack_get_version_string.argtypes = None
    jacklib.jack_get_version_string.restype  = c_char_p
except:
    jacklib.jack_get_version_string = None

try:
    jacklib.jack_client_open.argtypes = [c_char_p, jack_options_t, POINTER(jack_status_t), c_char_p]
    jacklib.jack_client_open.restype  = POINTER(jack_client_t)
except:
    jacklib.jack_client_open = None

try:
    jacklib.jack_client_rename.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_client_rename.restype  = c_char_p
except:
    jacklib.jack_client_rename = None

try:
    jacklib.jack_client_close.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_client_close.restype  = c_int
except:
    jacklib.jack_client_close = None

try:
    jacklib.jack_client_name_size.argtypes = None
    jacklib.jack_client_name_size.restype  = c_int
except:
    jacklib.jack_client_name_size = None

try:
    jacklib.jack_get_client_name.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_get_client_name.restype  = c_char_p
except:
    jacklib.jack_get_client_name = None

try:
    jacklib.jack_activate.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_activate.restype  = c_int
except:
    jacklib.jack_activate = None

try:
    jacklib.jack_deactivate.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_deactivate.restype  = c_int
except:
    jacklib.jack_deactivate = None

try:
    jacklib.jack_get_client_pid.argtypes = [c_char_p]
    jacklib.jack_get_client_pid.restype  = c_int
except:
    jacklib.jack_get_client_pid = None

try:
    jacklib.jack_is_realtime.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_is_realtime.restype  = c_int
except:
    jacklib.jack_is_realtime = None

def get_version_string(): # JACK2 only
    if jacklib.jack_get_version_string:
        return jacklib.jack_get_version_string()
    return None

def client_open(client_name, options, status, uuid=""):
    if jacklib.jack_client_open:
        return jacklib.jack_client_open(client_name.encode("utf-8"), options, status, uuid.encode("utf-8") if uuid else None)
    return None

def client_rename(client, new_name):
    if jacklib.jack_client_rename:
        return jacklib.jack_client_rename(client, new_name.encode("utf-8"))
    return None

def client_close(client):
    if jacklib.jack_client_close:
        return jacklib.jack_client_close(client)
    return -1

def client_name_size():
    if jacklib.jack_client_name_size:
        return jacklib.jack_client_name_size()
    return 0

def get_client_name(client):
    if jacklib.jack_get_client_name:
        return jacklib.jack_get_client_name(client)
    return None

def activate(client):
    if jacklib.jack_activate:
        return jacklib.jack_activate(client)
    return -1

def deactivate(client):
    if jacklib.jack_deactivate:
        return jacklib.jack_deactivate(client)
    return -1

def get_client_pid(name): # JACK2 only
    if jacklib.jack_get_client_pid:
        return jacklib.jack_get_client_pid(name.encode("utf-8"))
    return 0

def is_realtime(client):
    if jacklib.jack_is_realtime:
        return jacklib.jack_is_realtime(client)
    return 0

# ------------------------------------------------------------------------------------------------------------
# Non-Callback API

global _thread_callback
_thread_callback = None

try:
    jacklib.jack_cycle_wait.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_cycle_wait.restype  = jack_nframes_t
except:
    jacklib.jack_cycle_wait = None

try:
    jacklib.jack_cycle_signal.argtypes = [POINTER(jack_client_t), c_int]
    jacklib.jack_cycle_signal.restype  = None
except:
    jacklib.jack_cycle_signal = None

try:
    jacklib.jack_set_process_thread.argtypes = [POINTER(jack_client_t), JackThreadCallback, c_void_p]
    jacklib.jack_set_process_thread.restype  = c_int
except:
    jacklib.jack_set_process_thread = None

def cycle_wait(client):
    if jacklib.jack_cycle_wait:
        return jacklib.jack_cycle_wait(client)
    return 0

def cycle_signal(client, status):
    if jacklib.jack_cycle_signal:
        jacklib.jack_cycle_signal(client, status)

def set_process_thread(client, thread_callback, arg):
    if jacklib.jack_set_process_thread:
        global _thread_callback
        _thread_callback = JackThreadCallback(thread_callback)
        return jacklib.jack_set_process_thread(client, _thread_callback, arg)
    return -1

# ------------------------------------------------------------------------------------------------------------
# Client Callbacks

global _thread_init_callback
global _shutdown_callback
global _info_shutdown_callback
global _process_callback
global _freewheel_callback
global _bufsize_callback
global _srate_callback
global _client_registration_callback
global _client_rename_callback
global _port_registration_callback
global _port_connect_callback
global _port_rename_callback
global _graph_callback
global _xrun_callback
global _latency_callback

_thread_init_callback = _shutdown_callback = _info_shutdown_callback = None
_process_callback = _freewheel_callback = _bufsize_callback = _srate_callback = None
_client_registration_callback = _client_rename_callback = None
_port_registration_callback = _port_connect_callback = _port_rename_callback = None
_graph_callback = _xrun_callback = _latency_callback = None

try:
    jacklib.jack_set_thread_init_callback.argtypes = [POINTER(jack_client_t), JackThreadInitCallback, c_void_p]
    jacklib.jack_set_thread_init_callback.restype = c_int
except:
    jacklib.jack_set_thread_init_callback = None

try:
    jacklib.jack_on_shutdown.argtypes = [POINTER(jack_client_t), JackShutdownCallback, c_void_p]
    jacklib.jack_on_shutdown.restype = None
except:
    jacklib.jack_on_shutdown = None

try:
    jacklib.jack_on_info_shutdown.argtypes = [POINTER(jack_client_t), JackInfoShutdownCallback, c_void_p]
    jacklib.jack_on_info_shutdown.restype = None
except:
    jacklib.jack_on_info_shutdown = None

try:
    jacklib.jack_set_process_callback.argtypes = [POINTER(jack_client_t), JackProcessCallback, c_void_p]
    jacklib.jack_set_process_callback.restype = c_int
except:
    jacklib.jack_set_process_callback = None

try:
    jacklib.jack_set_freewheel_callback.argtypes = [POINTER(jack_client_t), JackFreewheelCallback, c_void_p]
    jacklib.jack_set_freewheel_callback.restype = c_int
except:
    jacklib.jack_set_freewheel_callback = None

try:
    jacklib.jack_set_buffer_size_callback.argtypes = [POINTER(jack_client_t), JackBufferSizeCallback, c_void_p]
    jacklib.jack_set_buffer_size_callback.restype = c_int
except:
    jacklib.jack_set_buffer_size_callback = None

try:
    jacklib.jack_set_sample_rate_callback.argtypes = [POINTER(jack_client_t), JackSampleRateCallback, c_void_p]
    jacklib.jack_set_sample_rate_callback.restype = c_int
except:
    jacklib.jack_set_sample_rate_callback = None

try:
    jacklib.jack_set_client_registration_callback.argtypes = [POINTER(jack_client_t), JackClientRegistrationCallback, c_void_p]
    jacklib.jack_set_client_registration_callback.restype = c_int
except:
    jacklib.jack_set_client_registration_callback = None

try:
    jacklib.jack_set_client_rename_callback.argtypes = [POINTER(jack_client_t), JackClientRenameCallback, c_void_p]
    jacklib.jack_set_client_rename_callback.restype = c_int
except:
    jacklib.jack_set_client_rename_callback = None

try:
    jacklib.jack_set_port_registration_callback.argtypes = [POINTER(jack_client_t), JackPortRegistrationCallback, c_void_p]
    jacklib.jack_set_port_registration_callback.restype = c_int
except:
    jacklib.jack_set_port_registration_callback = None

try:
    jacklib.jack_set_port_connect_callback.argtypes = [POINTER(jack_client_t), JackPortConnectCallback, c_void_p]
    jacklib.jack_set_port_connect_callback.restype = c_int
except:
    jacklib.jack_set_port_connect_callback = None

try:
    jacklib.jack_set_port_rename_callback.argtypes = [POINTER(jack_client_t), JackPortRenameCallback, c_void_p]
    jacklib.jack_set_port_rename_callback.restype = c_int
except:
    jacklib.jack_set_port_rename_callback = None

try:
    jacklib.jack_set_graph_order_callback.argtypes = [POINTER(jack_client_t), JackGraphOrderCallback, c_void_p]
    jacklib.jack_set_graph_order_callback.restype = c_int
except:
    jacklib.jack_set_graph_order_callback = None

try:
    jacklib.jack_set_xrun_callback.argtypes = [POINTER(jack_client_t), JackXRunCallback, c_void_p]
    jacklib.jack_set_xrun_callback.restype = c_int
except:
    jacklib.jack_set_xrun_callback = None

try:
    jacklib.jack_set_latency_callback.argtypes = [POINTER(jack_client_t), JackLatencyCallback, c_void_p]
    jacklib.jack_set_latency_callback.restype = c_int
except:
    jacklib.jack_set_latency_callback = None

def set_thread_init_callback(client, thread_init_callback, arg):
    if jacklib.jack_set_thread_init_callback:
        global _thread_init_callback
        _thread_init_callback = JackThreadInitCallback(thread_init_callback)
        return jacklib.jack_set_thread_init_callback(client, _thread_init_callback, arg)
    return -1

def on_shutdown(client, shutdown_callback, arg):
    if jacklib.jack_on_shutdown:
        global _shutdown_callback
        _shutdown_callback = JackShutdownCallback(shutdown_callback)
        jacklib.jack_on_shutdown(client, _shutdown_callback, arg)

def on_info_shutdown(client, info_shutdown_callback, arg):
    if jacklib.jack_on_info_shutdown:
        global _info_shutdown_callback
        _info_shutdown_callback = JackInfoShutdownCallback(info_shutdown_callback)
        jacklib.jack_on_info_shutdown(client, _info_shutdown_callback, arg)

def set_process_callback(client, process_callback, arg):
    if jacklib.jack_set_process_callback:
        global _process_callback
        _process_callback = JackProcessCallback(process_callback)
        return jacklib.jack_set_process_callback(client, _process_callback, arg)
    return -1

def set_freewheel_callback(client, freewheel_callback, arg):
    if jacklib.jack_set_freewheel_callback:
        global _freewheel_callback
        _freewheel_callback = JackFreewheelCallback(freewheel_callback)
        return jacklib.jack_set_freewheel_callback(client, _freewheel_callback, arg)
    return -1

def set_buffer_size_callback(client, bufsize_callback, arg):
    if jacklib.jack_set_buffer_size_callback:
        global _bufsize_callback
        _bufsize_callback = JackBufferSizeCallback(bufsize_callback)
        return jacklib.jack_set_buffer_size_callback(client, _bufsize_callback, arg)
    return -1

def set_sample_rate_callback(client, srate_callback, arg):
    if jacklib.jack_set_sample_rate_callback:
        global _srate_callback
        _srate_callback = JackSampleRateCallback(srate_callback)
        return jacklib.jack_set_sample_rate_callback(client, _srate_callback, arg)
    return -1

def set_client_registration_callback(client, client_registration_callback, arg):
    if jacklib.jack_set_client_registration_callback:
        global _client_registration_callback
        _client_registration_callback = JackClientRegistrationCallback(client_registration_callback)
        return jacklib.jack_set_client_registration_callback(client, _client_registration_callback, arg)
    return -1

def set_client_rename_callback(client, client_rename_callback, arg): # JACK2 only
    if jacklib.jack_set_client_rename_callback:
        global _client_rename_callback
        _client_rename_callback = JackClientRenameCallback(client_rename_callback)
        return jacklib.jack_set_client_rename_callback(client, _client_rename_callback, arg)
    return -1

def set_port_registration_callback(client, port_registration_callback, arg):
    if jacklib.jack_set_port_registration_callback:
        global _port_registration_callback
        _port_registration_callback = JackPortRegistrationCallback(port_registration_callback)
        return jacklib.jack_set_port_registration_callback(client, _port_registration_callback, arg)
    return -1

def set_port_connect_callback(client, connect_callback, arg):
    if jacklib.jack_set_port_connect_callback:
        global _port_connect_callback
        _port_connect_callback = JackPortConnectCallback(connect_callback)
        return jacklib.jack_set_port_connect_callback(client, _port_connect_callback, arg)
    return -1

def set_port_rename_callback(client, rename_callback, arg): # JACK2 only
    if jacklib.jack_set_port_rename_callback:
        global _port_rename_callback
        _port_rename_callback = JackPortRenameCallback(rename_callback)
        return jacklib.jack_set_port_rename_callback(client, _port_rename_callback, arg)
    return -1

def set_graph_order_callback(client, graph_callback, arg):
    if jacklib.jack_set_graph_order_callback:
        global _graph_callback
        _graph_callback = JackGraphOrderCallback(graph_callback)
        return jacklib.jack_set_graph_order_callback(client, _graph_callback, arg)
    return -1

def set_xrun_callback(client, xrun_callback, arg):
    if jacklib.jack_set_xrun_callback:
        global _xrun_callback
        _xrun_callback = JackXRunCallback(xrun_callback)
        return jacklib.jack_set_xrun_callback(client, _xrun_callback, arg)
    return -1

def set_latency_callback(client, latency_callback, arg):
    if jacklib.jack_set_latency_callback:
        global _latency_callback
        _latency_callback = JackLatencyCallback(latency_callback)
        return jacklib.jack_set_latency_callback(client, _latency_callback, arg)
    return -1

# ------------------------------------------------------------------------------------------------------------
# Server Control

jacklib.jack_set_freewheel.argtypes = [POINTER(jack_client_t), c_int]
jacklib.jack_set_freewheel.restype = c_int

jacklib.jack_set_buffer_size.argtypes = [POINTER(jack_client_t), jack_nframes_t]
jacklib.jack_set_buffer_size.restype = c_int

jacklib.jack_get_sample_rate.argtypes = [POINTER(jack_client_t)]
jacklib.jack_get_sample_rate.restype = jack_nframes_t

jacklib.jack_get_buffer_size.argtypes = [POINTER(jack_client_t)]
jacklib.jack_get_buffer_size.restype = jack_nframes_t

jacklib.jack_engine_takeover_timebase.argtypes = [POINTER(jack_client_t)]
jacklib.jack_engine_takeover_timebase.restype = c_int

jacklib.jack_cpu_load.argtypes = [POINTER(jack_client_t)]
jacklib.jack_cpu_load.restype = c_float

def set_freewheel(client, onoff):
    return jacklib.jack_set_freewheel(client, onoff)

def set_buffer_size(client, nframes):
    return jacklib.jack_set_buffer_size(client, nframes)

def get_sample_rate(client):
    return jacklib.jack_get_sample_rate(client)

def get_buffer_size(client):
    return jacklib.jack_get_buffer_size(client)

def engine_takeover_timebase(client):
    return jacklib.jack_engine_takeover_timebase(client)

def cpu_load(client):
    return jacklib.jack_cpu_load(client)

# ------------------------------------------------------------------------------------------------------------
# Port Functions

jacklib.jack_port_register.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p, c_ulong, c_ulong]
jacklib.jack_port_register.restype = POINTER(jack_port_t)

jacklib.jack_port_unregister.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_port_unregister.restype = c_int

jacklib.jack_port_get_buffer.argtypes = [POINTER(jack_port_t), jack_nframes_t]
jacklib.jack_port_get_buffer.restype = c_void_p

jacklib.jack_port_name.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_name.restype = c_char_p

jacklib.jack_port_short_name.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_short_name.restype = c_char_p

jacklib.jack_port_flags.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_flags.restype = c_int

jacklib.jack_port_type.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_type.restype = c_char_p

if JACK2:
    jacklib.jack_port_type_id.argtypes = [POINTER(jack_port_t)]
    jacklib.jack_port_type_id.restype = jack_port_type_id_t

jacklib.jack_port_is_mine.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_port_is_mine.restype = c_int

jacklib.jack_port_connected.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_connected.restype = c_int

jacklib.jack_port_connected_to.argtypes = [POINTER(jack_port_t), c_char_p]
jacklib.jack_port_connected_to.restype = c_int

jacklib.jack_port_get_connections.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_get_connections.restype = POINTER(c_char_p)

jacklib.jack_port_get_all_connections.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_port_get_all_connections.restype = POINTER(c_char_p)

jacklib.jack_port_tie.argtypes = [POINTER(jack_port_t), POINTER(jack_port_t)]
jacklib.jack_port_tie.restype = c_int

jacklib.jack_port_untie.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_untie.restype = c_int

jacklib.jack_port_set_name.argtypes = [POINTER(jack_port_t), c_char_p]
jacklib.jack_port_set_name.restype = c_int

jacklib.jack_port_set_alias.argtypes = [POINTER(jack_port_t), c_char_p]
jacklib.jack_port_set_alias.restype = c_int

jacklib.jack_port_unset_alias.argtypes = [POINTER(jack_port_t), c_char_p]
jacklib.jack_port_unset_alias.restype = c_int

jacklib.jack_port_get_aliases.argtypes = [POINTER(jack_port_t), POINTER(ARRAY(c_char_p, 2))]
jacklib.jack_port_get_aliases.restype = c_int

jacklib.jack_port_request_monitor.argtypes = [POINTER(jack_port_t), c_int]
jacklib.jack_port_request_monitor.restype = c_int

jacklib.jack_port_request_monitor_by_name.argtypes = [POINTER(jack_client_t), c_char_p, c_int]
jacklib.jack_port_request_monitor_by_name.restype = c_int

jacklib.jack_port_ensure_monitor.argtypes = [POINTER(jack_port_t), c_int]
jacklib.jack_port_ensure_monitor.restype = c_int

jacklib.jack_port_monitoring_input.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_monitoring_input.restype = c_int

jacklib.jack_connect.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p]
jacklib.jack_connect.restype = c_int

jacklib.jack_disconnect.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p]
jacklib.jack_disconnect.restype = c_int

jacklib.jack_port_disconnect.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_port_disconnect.restype = c_int

jacklib.jack_port_name_size.argtypes = None
jacklib.jack_port_name_size.restype = c_int

jacklib.jack_port_type_size.argtypes = None
jacklib.jack_port_type_size.restype = c_int

try:
    jacklib.jack_port_type_get_buffer_size.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_port_type_get_buffer_size.restype = c_size_t
except:
    jacklib.jack_port_type_get_buffer_size = None

def port_register(client, port_name, port_type, flags, buffer_size):
    return jacklib.jack_port_register(client, port_name.encode("utf-8"), port_type.encode("utf-8"), flags, buffer_size)

def port_unregister(client, port):
    return jacklib.jack_port_unregister(client, port)

def port_get_buffer(port, nframes):
    return jacklib.jack_port_get_buffer(port, nframes)

def port_name(port):
    return jacklib.jack_port_name(port)

def port_short_name(port):
    return jacklib.jack_port_short_name(port)

def port_flags(port):
    return jacklib.jack_port_flags(port)

def port_type(port):
    return jacklib.jack_port_type(port)

def port_type_id(port): # JACK2 only
    return jacklib.jack_port_type_id(port)

def port_is_mine(client, port):
    return jacklib.jack_port_is_mine(client, port)

def port_connected(port):
    return jacklib.jack_port_connected(port)

def port_connected_to(port, port_name):
    return jacklib.jack_port_connected_to(port, port_name.encode("utf-8"))

def port_get_connections(port):
    return jacklib.jack_port_get_connections(port)

def port_get_all_connections(client, port):
    return jacklib.jack_port_get_all_connections(client, port)

def port_tie(src, dst):
    return jacklib.jack_port_tie(src, dst)

def port_untie(port):
    return jacklib.jack_port_untie(port)

def port_set_name(port, port_name):
    return jacklib.jack_port_set_name(port, port_name.encode("utf-8"))

def port_set_alias(port, alias):
    return jacklib.jack_port_set_alias(port, alias.encode("utf-8"))

def port_unset_alias(port, alias):
    return jacklib.jack_port_unset_alias(port, alias.encode("utf-8"))

def port_get_aliases(port):
    # NOTE - this function has no 2nd argument in jacklib
    # Instead, aliases will be passed in return value, in form of (int ret, str alias1, str alias2)
    name_size = port_name_size()
    alias_type = c_char_p * 2
    aliases = alias_type(" ".encode("utf-8") * name_size, " ".encode("utf-8") * name_size)

    ret = jacklib.jack_port_get_aliases(port, pointer(aliases))
    return (ret, str(aliases[0], encoding="utf-8"), str(aliases[1], encoding="utf-8"))

def port_request_monitor(port, onoff):
    return jacklib.jack_port_request_monitor(port, onoff)

def port_request_monitor_by_name(client, port_name, onoff):
    return jacklib.jack_port_request_monitor_by_name(client, port_name.encode("utf-8"), onoff)

def port_ensure_monitor(port, onoff):
    return jacklib.jack_port_ensure_monitor(port, onoff)

def port_monitoring_input(port):
    return jacklib.jack_port_monitoring_input(port)

def connect(client, source_port, destination_port):
    return jacklib.jack_connect(client, source_port.encode("utf-8"), destination_port.encode("utf-8"))

def disconnect(client, source_port, destination_port):
    return jacklib.jack_disconnect(client, source_port.encode("utf-8"), destination_port.encode("utf-8"))

def port_disconnect(client, port):
    return jacklib.jack_port_disconnect(client, port)

def port_name_size():
    return jacklib.jack_port_name_size()

def port_type_size():
    return jacklib.jack_port_type_size()

def port_type_get_buffer_size(client, port_type):
    if jacklib.jack_port_type_get_buffer_size:
        return jacklib.jack_port_type_get_buffer_size(client, port_type.encode("utf-8"))
    return 0

# ------------------------------------------------------------------------------------------------------------
# Latency Functions

jacklib.jack_port_set_latency.argtypes = [POINTER(jack_port_t), jack_nframes_t]
jacklib.jack_port_set_latency.restype = None

try:
    jacklib.jack_port_get_latency_range.argtypes = [POINTER(jack_port_t), jack_latency_callback_mode_t, POINTER(jack_latency_range_t)]
    jacklib.jack_port_get_latency_range.restype = None
except:
    jacklib.jack_port_get_latency_range = None

try:
    jacklib.jack_port_set_latency_range.argtypes = [POINTER(jack_port_t), jack_latency_callback_mode_t, POINTER(jack_latency_range_t)]
    jacklib.jack_port_set_latency_range.restype = None
except:
    jacklib.jack_port_set_latency_range = None

jacklib.jack_recompute_total_latencies.argtypes = [POINTER(jack_client_t)]
jacklib.jack_recompute_total_latencies.restype = c_int

jacklib.jack_port_get_latency.argtypes = [POINTER(jack_port_t)]
jacklib.jack_port_get_latency.restype = jack_nframes_t

jacklib.jack_port_get_total_latency.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_port_get_total_latency.restype = jack_nframes_t

jacklib.jack_recompute_total_latency.argtypes = [POINTER(jack_client_t), POINTER(jack_port_t)]
jacklib.jack_recompute_total_latency.restype = c_int

def port_set_latency(port, nframes):
    jacklib.jack_port_set_latency(port, nframes)

def port_get_latency_range(port, mode, range_):
    if jacklib.jack_port_get_latency_range:
        jacklib.jack_port_get_latency_range(port, mode, range_)

def port_set_latency_range(port, mode, range_):
    if jacklib.jack_port_set_latency_range:
        jacklib.jack_port_set_latency_range(port, mode, range_)

def recompute_total_latencies():
    return jacklib.jack_recompute_total_latencies()

def port_get_latency(port):
    return jacklib.jack_port_get_latency(port)

def port_get_total_latency(client, port):
    return jacklib.jack_port_get_total_latency(client, port)

def recompute_total_latency(client, port):
    return jacklib.jack_recompute_total_latency(client, port)

# ------------------------------------------------------------------------------------------------------------
# Port Searching

jacklib.jack_get_ports.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p, c_ulong]
jacklib.jack_get_ports.restype = POINTER(c_char_p)

jacklib.jack_port_by_name.argtypes = [POINTER(jack_client_t), c_char_p]
jacklib.jack_port_by_name.restype = POINTER(jack_port_t)

jacklib.jack_port_by_id.argtypes = [POINTER(jack_client_t), jack_port_id_t]
jacklib.jack_port_by_id.restype = POINTER(jack_port_t)

def get_ports(client, port_name_pattern, type_name_pattern, flags):
    return jacklib.jack_get_ports(client, port_name_pattern.encode("utf-8"), type_name_pattern.encode("utf-8"), flags)

def port_by_name(client, port_name):
    return jacklib.jack_port_by_name(client, port_name.encode("utf-8"))

def port_by_id(client, port_id):
    return jacklib.jack_port_by_id(client, port_id)

# ------------------------------------------------------------------------------------------------------------
# Time Functions

jacklib.jack_frames_since_cycle_start.argtypes = [POINTER(jack_client_t)]
jacklib.jack_frames_since_cycle_start.restype = jack_nframes_t

jacklib.jack_frame_time.argtypes = [POINTER(jack_client_t)]
jacklib.jack_frame_time.restype = jack_nframes_t

jacklib.jack_last_frame_time.argtypes = [POINTER(jack_client_t)]
jacklib.jack_last_frame_time.restype = jack_nframes_t

try: # JACK_OPTIONAL_WEAK_EXPORT
    jacklib.jack_get_cycle_times.argtypes = [POINTER(jack_client_t), POINTER(jack_nframes_t), POINTER(jack_time_t), POINTER(jack_time_t), POINTER(c_float)]
    jacklib.jack_get_cycle_times.restype = c_int
except:
    jacklib.jack_get_cycle_times = None

jacklib.jack_frames_to_time.argtypes = [POINTER(jack_client_t), jack_nframes_t]
jacklib.jack_frames_to_time.restype = jack_time_t

jacklib.jack_time_to_frames.argtypes = [POINTER(jack_client_t), jack_time_t]
jacklib.jack_time_to_frames.restype = jack_nframes_t

jacklib.jack_get_time.argtypes = None
jacklib.jack_get_time.restype = jack_time_t

def frames_since_cycle_start(client):
    return jacklib.jack_frames_since_cycle_start(client)

def frame_time(client):
    return jacklib.jack_frame_time(client)

def last_frame_time(client):
    return jacklib.jack_last_frame_time(client)

def get_cycle_times(client, current_frames, current_usecs, next_usecs, period_usecs): # JACK_OPTIONAL_WEAK_EXPORT
    if jacklib.jack_frames_to_time:
        return jacklib.jack_get_cycle_times(client, current_frames, current_usecs, next_usecs, period_usecs)
    return -1

def frames_to_time(client, nframes):
    return jacklib.jack_frames_to_time(client, nframes)

def time_to_frames(client, time):
    return jacklib.jack_time_to_frames(client, time)

def get_time():
    return jacklib.jack_get_time()

# ------------------------------------------------------------------------------------------------------------
# Error Output

# TODO

# ------------------------------------------------------------------------------------------------------------
# Misc

jacklib.jack_free.argtypes = [c_void_p]
jacklib.jack_free.restype = None

def free(ptr):
    return jacklib.jack_free(ptr)

# ------------------------------------------------------------------------------------------------------------
# Transport

global _sync_callback
global _timebase_callback
_sync_callback = _timebase_callback = None

jacklib.jack_release_timebase.argtypes = [POINTER(jack_client_t)]
jacklib.jack_release_timebase.restype = c_int

jacklib.jack_set_sync_callback.argtypes = [POINTER(jack_client_t), JackSyncCallback, c_void_p]
jacklib.jack_set_sync_callback.restype = c_int

jacklib.jack_set_sync_timeout.argtypes = [POINTER(jack_client_t), jack_time_t]
jacklib.jack_set_sync_timeout.restype = c_int

jacklib.jack_set_timebase_callback.argtypes = [POINTER(jack_client_t), c_int, JackTimebaseCallback, c_void_p]
jacklib.jack_set_timebase_callback.restype = c_int

jacklib.jack_transport_locate.argtypes = [POINTER(jack_client_t), jack_nframes_t]
jacklib.jack_transport_locate.restype = c_int

jacklib.jack_transport_query.argtypes = [POINTER(jack_client_t), POINTER(jack_position_t)]
jacklib.jack_transport_query.restype = jack_transport_state_t

jacklib.jack_get_current_transport_frame.argtypes = [POINTER(jack_client_t)]
jacklib.jack_get_current_transport_frame.restype = jack_nframes_t

jacklib.jack_transport_reposition.argtypes = [POINTER(jack_client_t), POINTER(jack_position_t)]
jacklib.jack_transport_reposition.restype = c_int

jacklib.jack_transport_start.argtypes = [POINTER(jack_client_t)]
jacklib.jack_transport_start.restype = None

jacklib.jack_transport_stop.argtypes = [POINTER(jack_client_t)]
jacklib.jack_transport_stop.restype = None

def release_timebase(client):
    return jacklib.jack_release_timebase(client)

def set_sync_callback(client, sync_callback, arg):
    global _sync_callback
    _sync_callback = JackSyncCallback(sync_callback)
    return jacklib.jack_set_sync_callback(client, _sync_callback, arg)

def set_sync_timeout(client, timeout):
    return jacklib.jack_set_sync_timeout(client, timeout)

def set_timebase_callback(client, conditional, timebase_callback, arg):
    global _timebase_callback
    _timebase_callback = JackTimebaseCallback(timebase_callback)
    return jacklib.jack_set_timebase_callback(client, conditional, _timebase_callback, arg)

def transport_locate(client, frame):
    return jacklib.jack_transport_locate(client, frame)

def transport_query(client, pos):
    return jacklib.jack_transport_query(client, pos)

def get_current_transport_frame(client):
    return jacklib.jack_get_current_transport_frame(client)

def transport_reposition(client, pos):
    return jacklib.jack_transport_reposition(client, pos)

def transport_start(client):
    return jacklib.jack_transport_start(client)

def transport_stop(client):
    return jacklib.jack_transport_stop(client)

# ------------------------------------------------------------------------------------------------------------
# MIDI

jacklib.jack_midi_get_event_count.argtypes = [c_void_p]
jacklib.jack_midi_get_event_count.restype = jack_nframes_t

jacklib.jack_midi_event_get.argtypes = [POINTER(jack_midi_event_t), c_void_p, c_uint32]
jacklib.jack_midi_event_get.restype = c_int

jacklib.jack_midi_clear_buffer.argtypes = [c_void_p]
jacklib.jack_midi_clear_buffer.restype = None

jacklib.jack_midi_max_event_size.argtypes = [c_void_p]
jacklib.jack_midi_max_event_size.restype = c_size_t

jacklib.jack_midi_event_reserve.argtypes = [c_void_p, jack_nframes_t, c_size_t]
jacklib.jack_midi_event_reserve.restype = POINTER(jack_midi_data_t)

jacklib.jack_midi_event_write.argtypes = [c_void_p, jack_nframes_t, POINTER(jack_midi_data_t), c_size_t]
jacklib.jack_midi_event_write.restype = c_int

jacklib.jack_midi_get_lost_event_count.argtypes = [c_void_p]
jacklib.jack_midi_get_lost_event_count.restype = c_uint32

def midi_get_event_count(port_buffer):
    return jacklib.jack_midi_get_event_count(port_buffer)

def midi_event_get(event, port_buffer, event_index):
    return jacklib.jack_midi_event_get(event, port_buffer, event_index)

def midi_clear_buffer(port_buffer):
    return jacklib.jack_midi_clear_buffer(port_buffer)

def midi_max_event_size(port_buffer):
    return jacklib.jack_midi_max_event_size(port_buffer)

def midi_event_reserve(port_buffer, time, data_size):
    return jacklib.jack_midi_event_reserve(port_buffer, time, data_size)

def midi_event_write(port_buffer, time, data, data_size):
    return jacklib.jack_midi_event_write(port_buffer, time, data, data_size)

def midi_get_lost_event_count(port_buffer):
    return jacklib.jack_midi_get_lost_event_count(port_buffer)

# ------------------------------------------------------------------------------------------------------------
# Session

global _session_callback
_session_callback = None

try:
    jacklib.jack_set_session_callback.argtypes = [POINTER(jack_client_t), JackSessionCallback, c_void_p]
    jacklib.jack_set_session_callback.restype = c_int
except:
    jacklib.jack_set_session_callback = None

try:
    jacklib.jack_session_reply.argtypes = [POINTER(jack_client_t), POINTER(jack_session_event_t)]
    jacklib.jack_session_reply.restype = c_int
except:
    jacklib.jack_session_reply = None

try:
    jacklib.jack_session_event_free.argtypes = [POINTER(jack_session_event_t)]
    jacklib.jack_session_event_free.restype = None
except:
    jacklib.jack_session_event_free = None

try:
    jacklib.jack_client_get_uuid.argtypes = [POINTER(jack_client_t)]
    jacklib.jack_client_get_uuid.restype = c_char_p
except:
    jacklib.jack_client_get_uuid = None

try:
    jacklib.jack_session_notify.argtypes = [POINTER(jack_client_t), c_char_p, jack_session_event_type_t, c_char_p]
    jacklib.jack_session_notify.restype = POINTER(jack_session_command_t)
except:
    jacklib.jack_session_notify = None

try:
    jacklib.jack_session_commands_free.argtypes = [POINTER(jack_session_command_t)]
    jacklib.jack_session_commands_free.restype = None
except:
    jacklib.jack_session_commands_free = None

try:
    jacklib.jack_get_uuid_for_client_name.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_get_uuid_for_client_name.restype = c_char_p
except:
    jacklib.jack_get_uuid_for_client_name = None

try:
    jacklib.jack_get_client_name_by_uuid.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_get_client_name_by_uuid.restype = c_char_p
except:
    jacklib.jack_get_client_name_by_uuid = None

try:
    jacklib.jack_reserve_client_name.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p]
    jacklib.jack_reserve_client_name.restype = c_int
except:
    jacklib.jack_reserve_client_name = None

try:
    jacklib.jack_client_has_session_callback.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_client_has_session_callback.restype = c_int
except:
    jacklib.jack_client_has_session_callback = None

def set_session_callback(client, session_callback, arg):
    if jacklib.jack_set_session_callback:
        global _session_callback
        _session_callback = JackSessionCallback(session_callback)
        return jacklib.jack_set_session_callback(client, _session_callback, arg)
    return -1

def session_reply(client, event):
    if jacklib.jack_session_reply:
        return jacklib.jack_session_reply(client, event)
    return -1

def session_event_free(event):
    if jacklib.jack_session_event_free:
        jacklib.jack_session_event_free(event)

def client_get_uuid(client):
    if jacklib.jack_client_get_uuid:
        return jacklib.jack_client_get_uuid(client)
    return None

def session_notify(client, target, type_, path):
    if jacklib.jack_session_notify:
        return jacklib.jack_session_notify(client, target.encode("utf-8"), type_, path.encode("utf-8"))
    return jack_session_command_t()

def session_commands_free(cmds):
    if jacklib.jack_session_commands_free:
        jacklib.jack_session_commands_free(cmds)

def get_uuid_for_client_name(client, client_name):
    if jacklib.jack_get_uuid_for_client_name:
        return jacklib.jack_get_uuid_for_client_name(client, client_name.encode("utf-8"))
    return None

def get_client_name_by_uuid(client, client_uuid):
    if jacklib.jack_get_client_name_by_uuid:
        return jacklib.jack_get_client_name_by_uuid(client, client_uuid.encode("utf-8"))
    return None

def reserve_client_name(client, name, uuid):
    if jacklib.jack_reserve_client_name:
        return jacklib.jack_reserve_client_name(client, name.encode("utf-8"), uuid.encode("utf-8"))
    return -1

def client_has_session_callback(client, client_name):
    if jacklib.jack_client_has_session_callback:
        return jacklib.jack_client_has_session_callback(client, client_name.encode("utf-8"))
    return -1

# ------------------------------------------------------------------------------------------------------------
# Custom

global _custom_appearance_callback
_custom_appearance_callback = None

try:
    jacklib.jack_custom_publish_data.argtypes = [POINTER(jack_client_t), c_char_p, c_void_p, c_size_t]
    jacklib.jack_custom_publish_data.restype  = c_int
except:
    jacklib.jack_custom_publish_data = None

try:
    jacklib.jack_custom_get_data.argtypes = [POINTER(jack_client_t), c_char_p, c_char_p, POINTER(c_void_p), POINTER(c_size_t)]
    jacklib.jack_custom_get_data.restype  = c_int
except:
    jacklib.jack_custom_get_data = None

try:
    jacklib.jack_custom_unpublish_data.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_custom_unpublish_data.restype  = c_int
except:
    jacklib.jack_custom_unpublish_data = None

try:
    jacklib.jack_custom_get_keys.argtypes = [POINTER(jack_client_t), c_char_p]
    jacklib.jack_custom_get_keys.restype  = POINTER(c_char_p)
except:
    jacklib.jack_custom_get_keys = None

try:
    jacklib.jack_custom_set_data_appearance_callback.argtypes = [POINTER(jack_client_t), JackCustomDataAppearanceCallback, c_void_p]
    jacklib.jack_custom_set_data_appearance_callback.restype  = c_int
except:
    jacklib.jack_custom_set_data_appearance_callback = None

def custom_publish_data(client, key, data, size):
    if jacklib.jack_custom_publish_data:
        return jacklib.jack_custom_publish_data(client, key.encode("utf-8"), data, size)
    return -1

def custom_get_data(client, client_name, key):
    # NOTE - this function has no extra arguments in jacklib
    # Instead, data and size will be passed in return value, in form of (int ret, void* data, size_t size)

    if jacklib.jack_custom_get_data:
        data = c_void_p(0)
        size = c_size_t(0)
        ret  = jacklib.jack_custom_get_data(client, client_name.encode("utf-8"), key.encode("utf-8"), pointer(data), pointer(size))
        return (ret, data, size)

    return (-1, None, 0)

def custom_unpublish_data(client, key):
    if jacklib.jack_custom_unpublish_data:
        return jacklib.jack_custom_unpublish_data(client, key.encode("utf-8"))
    return -1

def custom_get_keys(client, client_name):
    if jacklib.jack_custom_get_keys:
        return jacklib.jack_custom_get_keys(client, client_name.encode("utf-8"))
    return None

def custom_set_data_appearance_callback(client, custom_callback, arg):
    if jacklib.jack_custom_set_data_appearance_callback:
        global _custom_appearance_callback
        _custom_appearance_callback = JackCustomDataAppearanceCallback(custom_callback)
        return jacklib.jack_custom_set_data_appearance_callback(client, _custom_appearance_callback, arg)
    return -1



# -*- coding: utf-8 -*-

import os
import pathlib
import sys

from PyQt5.QtCore import QSettings, QDataStream, QIODevice, QUrl, QByteArray
from PyQt5.QtXml  import QDomDocument, QDomText
import ray
from daemon_tools import getAppConfigPath

QFileDialogMagic = 190

class PickerType:
    def __init__(self, config_path):
        self.config_path = config_path
        self.written     = False
        
    def makeBookmark(self, spath):
        pass
    
    def removeBookmark(self, spath):
        pass
    
    def getContents(self):
        if os.path.exists(self.config_path):
            try:
                file = open(self.config_path, 'r')
                contents = file.read()
                file.close()
                return contents
            except:
                return None
        else:
            return ""
        
    def printContents(self, contents):
        try:
            file = open(self.config_path, 'w')
        except:
            return False
        
        file.write(contents)
        file.close()
        
        return True
    
class PickerTypeGtk(PickerType):
    def makeBookmark(self, spath):
        if self.written:
            return
        
        url = pathlib.Path(spath).as_uri()
        
        config_dir = os.path.dirname(self.config_path)
        
        if not os.path.exists(config_dir):
            try:
                os.makedirs(config_dir)
            except:
                return
        
        contents = self.getContents()
        if contents == None:
            return
        
        bookmarks = contents.split('\n')
        
        if url in bookmarks:
            return
            
        contents += "%s\n" % url
        
        if self.printContents(contents):
            self.written = True
        
    def removeBookmark(self, spath):
        if not self.written:
            return
        
        url = pathlib.Path(spath).as_uri()
        
        contents = self.getContents()        
        if not contents:
            self.written = False
            return
        
        bookmarks = contents.split('\n')
        
        if url in bookmarks:
            bookmarks.remove(url)
        else:
            self.written = False
            return
            
        contents = ''
        for url in bookmarks:
            if url:
                contents += "%s\n" % url
        
        self.printContents(contents)
        self.written = False
        
class PickerTypeFltk(PickerType):
    def makeBookmark(self, spath):
        if self.written:
            return
        
        contents = self.getContents()
        if not contents:
            #we won't write a file for fltk if file doesn't already exists
            return
        
        lines = contents.split('\n')
        contents = ""
        num = -1
        empty_fav = False
        fav0_found = False
        
        for line in lines:
            if line.startswith('favorite') and ':' in line:
                fav0_found = True
                
                if line.partition(':')[2]:
                    if line.partition(':')[2] == spath:
                        #bookmark already written, do nothing
                        return
                    
                    num_s = line.partition(':')[0].replace('favorite', '', 1)
                    if num_s.isdigit():
                        num = int(num_s)
                else:
                    if not empty_fav:
                        line += spath
                        empty_fav = True
                
            if line or not fav0_found:
                contents+= "%s\n" % line
        
        if not empty_fav:
            num+=1
            contents += "favorite%.2d:%s" % (num, spath)
            
        if self.printContents(contents):
            self.written = True
            
    def removeBookmark(self, spath):
        if not self.written:
            return
           
        contents = self.getContents()
        if not contents:
            self.written = False
            return
        
        lines = contents.split('\n')
        favorites = []
        
        for line in lines:
            if line.startswith('favorite') and ':' in line:
                fav = line.partition(':')[2]
                favorites.append(fav)
        
        if not spath in favorites:
            self.written = False
            return
        
        favorites.remove(spath)
        contents = ""
        num = 0
        
        for line in lines:
            if line.startswith('favorite') and ':' in line:
                fav = ''
                if num < len(favorites):
                    fav = favorites[num]
                
                contents+= "favorite%.2d:%s\n" % (num, fav)
                num+=1
            else:
                contents+= "%s\n" % line
                    
        self.printContents(contents)
        self.written = False

class PickerTypeQt4(PickerType):
    def makeBookmark(self, spath):
        if self.written:
            return
        
        if not os.path.exists(self.config_path):
            #do not write shortcuts if file was not created by Qt4 himself
            return
        
        url = pathlib.Path(spath).as_uri()
        
        settings_qt4 = QSettings(self.config_path, QSettings.IniFormat)
        if not settings_qt4.isWritable():
            return
        
        data = settings_qt4.value('Qt/filedialog')
        stream = QDataStream(data, QIODevice.ReadOnly)
        
        magic   = stream.readUInt32()
        version = stream.readUInt32()
        if not (magic == QFileDialogMagic and version == 3):
            return
        
        split_states = stream.readBytes()
        
        bookmarks_len = stream.readUInt32()
        bookmarks = []
        for bm in range(bookmarks_len):
            qUrl = QUrl()
            stream >> qUrl
            
            if qUrl.isLocalFile() and qUrl.toLocalFile() == spath:
                #spath already in qt4 bookmarks
                return
            
            bookmarks.append(qUrl)
            
            
        history_len = stream.readUInt32()
        history = []
        for h in range(history_len):
            his = stream.readQString()
            history.append(his)
            
        current_dir = stream.readQString()
        header_data = stream.readBytes()
        view_mode   = stream.readUInt32()
        
        
        #now rewrite bytes
        
        new_data = QByteArray()
        new_stream = QDataStream(new_data, QIODevice.WriteOnly)
        
        new_stream.writeUInt32(magic)
        new_stream.writeUInt32(3)
        new_stream.writeBytes(split_states)
        new_stream.writeUInt32(bookmarks_len+1)
        for bm in bookmarks:
            new_stream << bm
            
        qUrl = QUrl(url)
        new_stream << qUrl
        
        new_stream.writeQStringList(history)
        new_stream.writeQString(current_dir)
        new_stream.writeBytes(header_data)
        new_stream.writeUInt32(view_mode)
        
        settings_qt4.setValue('Qt/filedialog', new_data)
        settings_qt4.sync()
        
        self.written = True
    
    def removeBookmark(self, spath):
        if not self.written:
            return
        
        if not os.path.exists(self.config_path):
            self.written = False
            return
        
        url = pathlib.Path(spath).as_uri()
        
        settings_qt4 = QSettings(self.config_path, QSettings.IniFormat)
        if not settings_qt4.isWritable():
            self.written = False
            return
        
        data = settings_qt4.value('Qt/filedialog')
        stream = QDataStream(data, QIODevice.ReadOnly)
        
        magic   = stream.readUInt32()
        version = stream.readUInt32()
        if not (magic == QFileDialogMagic and version == 3):
            self.written = False
            return
        
        split_states = stream.readBytes()
        
        bookmark_found = False
        bookmarks_len = stream.readUInt32()
        bookmarks = []
        for bm in range(bookmarks_len):
            qUrl = QUrl()
            stream >> qUrl
            
            if qUrl.isLocalFile() and qUrl.toLocalFile() == spath:
                bookmark_found = True
            else:
                bookmarks.append(qUrl)
                
        if not bookmark_found:
            self.written = False
            return
        
        history_len = stream.readUInt32()
        history = []
        for h in range(history_len):
            his = stream.readQString()
            history.append(his)
            
        current_dir = stream.readQString()
        header_data = stream.readBytes()
        view_mode   = stream.readUInt32()
        
        #now rewrite bytes
        
        new_data = QByteArray()
        new_stream = QDataStream(new_data, QIODevice.WriteOnly)
        
        new_stream.writeUInt32(magic)
        new_stream.writeUInt32(3)
        new_stream.writeBytes(split_states)
        new_stream.writeUInt32(bookmarks_len-1)
        for bm in bookmarks:
            new_stream << bm
            
        qUrl = QUrl(url)
        new_stream << qUrl
        
        new_stream.writeQStringList(history)
        new_stream.writeQString(current_dir)
        new_stream.writeBytes(header_data)
        new_stream.writeUInt32(view_mode)
        
        settings_qt4.setValue('Qt/filedialog', new_data)
        settings_qt4.sync()
        
        self.written = False

class PickerTypeQt5(PickerType):
    def makeBookmark(self, spath):
        if self.written:
            return
        
        if not os.path.exists(self.config_path):
            #do not write shortcuts if file was not created by Qt5 himself
            return
        
        url = pathlib.Path(spath).as_uri()
        
        settings_qt5 = QSettings(self.config_path, QSettings.IniFormat)
        if not settings_qt5.isWritable():
            return
        
        shortcuts = ray.getListInSettings(settings_qt5, 'FileDialog/shortcuts')
        
        for sc in shortcuts:
            sc_url = QUrl(sc)
            if sc_url.isLocalFile() and sc_url.toLocalFile() == spath:
                return
        
        shortcuts.append(url)
        
        settings_qt5.setValue('FileDialog/shortcuts', shortcuts)
        settings_qt5.sync()
        self.written = True
            
    def removeBookmark(self, spath):
        if not self.written:
            return
        
        if not os.path.exists(self.config_path):
            self.written = False
            return
        
        url = pathlib.Path(spath).as_uri()
        
        settings_qt5 = QSettings(self.config_path, QSettings.IniFormat)
        shortcuts = ray.getListInSettings(settings_qt5, 'FileDialog/shortcuts')
        
        for sc in shortcuts:
            sc_url = QUrl(sc)
            if sc_url.isLocalFile() and sc_url.toLocalFile() == spath:
                shortcuts.remove(sc)
                break
        else:
            self.written = False
            return
        
        settings_qt5.setValue('FileDialog/shortcuts', shortcuts)
        settings_qt5.sync()
        self.written = False

class PickerTypeKde5(PickerType):
    def makeBookmark(self, spath):
        if self.written:
            return
        
        contents = self.getContents()
        if not contents:
            #we won't write a file for kde5 if file doesn't already exists
            return
        
        url = pathlib.Path(spath).as_uri()
        
        xml = QDomDocument()
        xml.setContent(contents)
        content = xml.documentElement()
        if content.tagName() != 'xbel':
            return
        
        node = content.firstChild()
        while not node.isNull():
            el = node.toElement()
            if el.tagName() == 'bookmark':
                if el.attribute('href') == url:
                    #bookmark already exists
                    return
            
            node = node.nextSibling()
        
        bk = xml.createElement('bookmark')
        bk.setAttribute('href', url)
        title = xml.createElement('title')
        title_text = xml.createTextNode(os.path.basename(spath))
        title.appendChild(title_text)
        bk.appendChild(title)
        content.appendChild(bk)
        
        if self.printContents(xml.toString()):
            self.written = True
            
            
    def removeBookmark(self, spath):
        if not self.written:
            return
        
        contents = self.getContents()
        if not contents:
            self.written = False
            return
        
        url = pathlib.Path(spath).as_uri()
        
        xml = QDomDocument()
        xml.setContent(contents)
        content = xml.documentElement()
        if content.tagName() != 'xbel':
            self.written = False
            return
        
        node = content.firstChild()
        while not node.isNull():
            el = node.toElement()
            if el.tagName() == 'bookmark':
                if el.attribute('href') == url:
                    content.removeChild(node)
                    break
            
            node = node.nextSibling()
        else:
            self.written = False
            return
        
        self.printContents(xml.toString())
        self.written = False
        
        
class BookMarker:
    def __init__(self):
        self.bookmarks_memory = "%s/bookmarks.xml" % getAppConfigPath()
        self.daemon_port      = 0
        
        HOME = os.getenv('HOME')
        
        self.gtk2 = PickerTypeGtk("%s/.gtk-bookmarks" % HOME)
        self.gtk3 = PickerTypeGtk("%s/.config/gtk-3.0/bookmarks" % HOME)
        self.fltk = PickerTypeFltk("%s/.fltk/fltk.org/filechooser.prefs"
                                    % HOME)
        self.kde5 = PickerTypeKde5("%s/.local/share/user-places.xbel" % HOME)
        self.qt4  = PickerTypeQt4("%s/.config/Trolltech.conf" % HOME)
        self.qt5  = PickerTypeQt5("%s/.config/QtProject.conf" % HOME)
        
    def setDaemonPort(self, port):
        self.daemon_port = port
    
    def getXml(self):
        xml = QDomDocument()
        file_exists = False
        
        if os.path.exists(self.bookmarks_memory):
            try:
                file = open(self.bookmarks_memory, 'r')
                xml.setContent(file.read())
                file_exists = True
            except:
                try:
                    os.path.remove(self.bookmarks_memory)
                except:
                    return None
        
        if not file_exists:
            bms_xml = xml.createElement('Bookmarks')
            xml.appendChild(bms_xml)
            
        return xml
    
    def writeXmlFile(self, xml):
        try:
            file = open(self.bookmarks_memory, 'w')
            file.write(xml.toString())
        except:
            return
    
    def getPickersForXml(self):
        string = ":"
        if self.gtk2.written:
            string += "gtk2:"
        if self.gtk3.written:
            string += "gtk3:"
        if self.fltk.written:
            string += "fltk:"
        if self.kde5.written:
            string += "kde5:"
        if self.qt4.written:
            string += "qt4:"
        if self.qt5.written:
            string += "qt5:"
        
        return string
    
    def makeAll(self, spath):
        for picker in (self.gtk2, self.gtk3, self.fltk,
                       self.kde5, self.qt4, self.qt5):
            picker.makeBookmark(spath)
        
        xml = self.getXml()
        if not xml:
            return
            
        xml_content = xml.documentElement()
        node = xml_content.firstChild()
        
        bke = xml.createElement('bookmarker')
        bke.setAttribute('port', self.daemon_port)
        bke.setAttribute('session_path', spath)
        bke.setAttribute('pickers', self.getPickersForXml())
        node = xml_content.firstChild()
        xml_content.appendChild(bke)
        
        self.writeXmlFile(xml)
        
    def removeAll(self, spath):
        for picker in (self.gtk2, self.gtk3, self.fltk,
                       self.kde5, self.qt4, self.qt5):
            picker.removeBookmark(spath)
            
        xml = self.getXml()
        if not xml:
            return
            
        xml_content = xml.documentElement()
        nodes = xml_content.childNodes()
        for i in range(nodes.count()):
            node = nodes.at(i)
            
            bke = node.toElement()
            port = bke.attribute('port')
            session_path = bke.attribute('session_path')
            
            if (port.isdigit()
                    and int(port) == self.daemon_port
                    and session_path == spath):
                xml_content.removeChild(node)
                break
            
        self.writeXmlFile(xml)
            
    
    def clean(self, all_session_paths):
        xml = self.getXml()
        if not xml:
            return
        
        xml_content = xml.documentElement()
        nodes = xml_content.childNodes()
        nodes_to_remove = []
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            bke = node.toElement()
            spath   = bke.attribute('session_path')
            pickers = bke.attribute('pickers')
            
            if not spath:
                nodes_to_remove.append(node)
                continue
            
            if not spath in all_session_paths:
                if ":gtk2:" in pickers:
                    self.gtk2.written = True
                    self.gtk2.removeBookmark(spath)
                if ":gtk3:" in pickers:
                    self.gtk3.written = True
                    self.gtk3.removeBookmark(spath)
                if ":fltk:" in pickers:
                    self.fltk.written = True
                    self.fltk.removeBookmark(spath)
                if ":kde5:" in pickers:
                    self.kde5.written = True
                    self.kde5.removeBookmark(spath)
                if ":qt4:" in pickers:
                    self.qt4.written = True
                    self.qt4.removeBookmark(spath)
                if ":qt5:" in pickers:
                    self.qt5.written = True
                    self.qt5.removeBookmark(spath)
            
                nodes_to_remove.append(node)
            
        for node in nodes_to_remove:
            xml_content.removeChild(node)
            
        self.writeXmlFile(xml)
    
if __name__ == '__main__':
    bm_maker = BookMarker()
    bm_maker.makeAll(sys.argv[1])
    
        
        



from PyQt5.QtCore import QObject, pyqtSignal

instance = None

class Signaler(QObject):
    osc_recv = pyqtSignal(str, list, str, object)
    dummy_load_and_template = pyqtSignal(str, str, str)
    
    @staticmethod
    def instance():
        global instance
        
        if not instance:
            instance = Signaler()
        return instance
    
    def __init__(self):
        QObject.__init__(self)
        global instance
        instance = self





from PyQt5.QtCore import Qt, QDateTime, QDate, QTime
from PyQt5.QtWidgets import QDialogButtonBox, QTreeWidgetItem

import ray

from child_dialogs import ChildDialog
from gui_tools import _translate, RS

import ui_list_snapshots
import ui_snapshot_name
import ui_snapshots_info

GROUP_ELEMENT = 0
GROUP_DAY     = 1
GROUP_MONTH   = 2
GROUP_YEAR    = 3
GROUP_MAIN    = 4

class Snapshot:
    valid = False
    text  = ''
    sub_type = GROUP_ELEMENT
    item = None
    before_rewind_to = ''
    date_time = None
    rewind_date_time = None
    session_name = ""
    label = ''
    rewind_label = ''
    ref = ''
    
    def __init__(self, date_time):
        self.date_time = date_time
    
    def __lt__(self, other):
        if not other.isValid():
            return True
        
        if not self.isValid():
            return False
        
        return self.date_time < other.date_time
    
    def year(self):
        return self.date_time.date().year()
    
    def month(self):
        return self.date_time.date().month()
    
    def day(self):
        return self.date_time.date().day()
    
    def isValid(self):
        if not self.date_time:
            return False
        
        return self.date_time.isValid()
    
    def isToday(self):
        if not self.date_time:
            return False
        
        return bool(self.date_time.date() == QDate.currentDate())
    
    def isYesterday(self):
        if not self.date_time:
            return False
        
        return bool(self.date_time.date() == QDate.currentDate().addDays(-1))
    
    def canTake(self, other):
        return False
    
    def reOrganize(self):
        pass
    
    def add(self):
        pass
    
    def commonGroup(self, other):
        if not (self.isValid() and other.isValid()):
            return GROUP_MAIN
        
        common_group = GROUP_MAIN
        
        if self.year() == other.year():
            common_group = GROUP_YEAR
            if self.month() == other.month():
                common_group = GROUP_MONTH
                if self.day() == other.day():
                    common_group = GROUP_DAY
        
        if common_group <= self.sub_type:
            return self.sub_type +1
        
        return common_group
    
    
    
    def makeItem(self, sub_type):
        if self.isToday():
            day_string = _translate('snapshots', 'Today')
        elif self.isYesterday():
            day_string = _translate('snapshots', 'Yesterday')
        elif self.isValid():
            day_string = self.date_time.toString('dddd d MMMM yyyy')
        
        
        if not self.isValid():
            display_text = self.text
        else:
            display_text = _translate('snapshots', "%s at %s") % (
                            day_string,
                            self.date_time.toString('HH:mm'))
            
            if sub_type in (GROUP_YEAR, GROUP_MONTH):
                if not self.isToday() or self.isYesterday():
                    day_string = self.date_time.toString('dddd d MMMM')
                    
                display_text = _translate('snapshots', "%s at %s") % (
                                    day_string,
                                    self.date_time.toString('HH:mm'))
                
            elif sub_type == GROUP_DAY:
                display_text = _translate('snapshots', "at %s") \
                                 % self.date_time.toString('HH:mm')
            
            if self.rewind_date_time:
                display_text += '\n'
                display_text += _translate('snapshots', "before rewind to ")
                
                if self.rewind_label:
                    display_text += self.rewind_label
                elif self.rewind_date_time.date() == self.date_time.date():
                    display_text += self.rewind_date_time.toString('hh:mm')
                elif (self.rewind_date_time.date().year()
                    == self.date_time.date().year()):
                    display_text += self.rewind_date_time.toString('d MMM hh:mm')
                else:
                    display_text += self.rewind_date_time.toString('d MMM yyyy hh:mm')
                    
            elif self.session_name:
                display_text += "\nsession name: %s" % self.session_name
        
        if self.label:
            display_text += "\n%s" % self.label
        
        item = QTreeWidgetItem([display_text])
        item.setData(0, Qt.UserRole, self.text)
        
        return item


class SnapGroup(Snapshot):
    def __init__(self, date_time=None, sub_type=GROUP_MAIN):
        Snapshot.__init__(self, date_time)
        self.sub_type = sub_type
        self.valid = True
        self.snapshots = []
    
    def canTake(self, other):
        if self.sub_type <= other.sub_type:
            return False
        
        if self.sub_type == GROUP_MAIN:
            return True
        
        if self.year() != other.year():
            return False
        
        if self.sub_type == GROUP_YEAR:
            return True
        
        if self.month() != other.month():
            return False
        
        if self.sub_type == GROUP_MONTH:
            return True
        
        if self.day() != other.day():
            return False
        
        return True
    
    def add(self, new_snapshot):
        if not new_snapshot.isValid():
            self.snapshots.append(new_snapshot)
            return
        
        if self.sub_type <= 1:
            # If this group (self) is a day group, just add this snapshot
            self.snapshots.append(new_snapshot)
            return
        
        for snapshot in self.snapshots:
            if snapshot.canTake(new_snapshot):
                # if a snapgroup can take this snapshot,
                # just add this snapshot to this snapgroup.
                snapshot.add(new_snapshot)
                return
        
        smallest_cg = self.sub_type
        
        # find the smallest common group with any other 
        for snapshot in self.snapshots:
            common_group = snapshot.commonGroup(new_snapshot)
            if common_group < smallest_cg:
                smallest_cg = common_group
        
        # check if there are snaps not common
        # with the smallest common group find above (smallest_cg)
        for snapshot in self.snapshots:
            common_group = snapshot.commonGroup(new_snapshot)
            if common_group != smallest_cg:
                break
        else:
            # There is no snap outside of smallest_cg
            # but there are maybe others snapshots to group together
            cg_final = 0
            compare_snap = Snapshot(None)
            
            for cg in (GROUP_DAY, GROUP_MONTH, GROUP_YEAR):
                if cg_final:
                    break
                
                if cg >= smallest_cg:
                    continue
                
                # compare all existing snapshots with all others
                for i in range(len(self.snapshots)):
                    if cg_final:
                        break
                    
                    compare_snap = self.snapshots[i]
                    if compare_snap.sub_type >= cg:
                        continue
                    
                    for j in range(len(self.snapshots)):
                        if j <= i:
                            # prevent compare to itself or already compared
                            continue
                        
                        snapshot = self.snapshots[j]
                        if snapshot.sub_type >= cg:
                            continue
                        
                        if (snapshot.commonGroup(compare_snap) == cg
                                and snapshot.commonGroup(new_snapshot) > cg):
                            cg_final = cg
                            break
            
            if cg_final:
                snap_group = SnapGroup(compare_snap.date_time, cg_final)
                self.addGroup(snap_group)
                
            self.snapshots.append(new_snapshot)
            return
        
        # create group and add to this all snaps which have to.
        snap_group = SnapGroup(new_snapshot.date_time, smallest_cg)
        snap_group.add(new_snapshot)
        self.addGroup(snap_group)
    
    def addGroup(self, snap_group):
        to_rem = []
        
        for i in range(len(self.snapshots)):
            snapshot = self.snapshots[i]
            if snap_group.canTake(snapshot):
                snap_group.add(snapshot)
                to_rem.append(i)
                
        to_rem.reverse()
        for i in to_rem:
            self.snapshots.__delitem__(i)
        
        self.snapshots.append(snap_group)
    
    def sort(self):
        for snapshot in self.snapshots:
            if snapshot.sub_type:
                snapshot.sort()
                
        self.snapshots.sort()
        self.snapshots.reverse()
                    
    def makeItem(self, sub_type=GROUP_MAIN):
        display_text = ''
        
        if self.sub_type == GROUP_MAIN:
            return None
        
        if not self.date_time:
            display_text = self.text
        elif self.sub_type == GROUP_YEAR:
            display_text = self.date_time.toString('yyyy')
        elif self.sub_type == GROUP_MONTH:
            display_text = self.date_time.toString('MMMM yyyy')
        elif self.sub_type == GROUP_DAY:
            display_text = self.date_time.toString('dddd d MMMM yyyy')
            if self.isToday():
                display_text = _translate('snapshots', 'Today')
            elif self.isYesterday():
                display_text = _translate('snapshots', 'Yesterday')
            
        item = QTreeWidgetItem([display_text])
        
        for snapshot in self.snapshots:
            sub_item = snapshot.makeItem(self.sub_type)
            item.addChild(sub_item)
        
        # set this group item not selectable
        item.setFlags(item.flags() & ~Qt.ItemIsSelectable)
        
        return item

class TakeSnapshotDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_snapshot_name.Ui_Dialog()
        self.ui.setupUi(self)
        
        self.ui.lineEdit.textChanged.connect(self.textChanged)
        #self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.pushButtonSave.setEnabled(False)
        self.ui.pushButtonSnapshot.setEnabled(False)
        
        self.__save_asked = False
        self.ui.pushButtonSave.clicked.connect(self.acceptWithSave)
        
    def textChanged(self, text):
        #self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(bool(text))
        self.ui.pushButtonSave.setEnabled(bool(text))
        self.ui.pushButtonSnapshot.setEnabled(bool(text))
        
    def getSnapshotName(self):
        return self.ui.lineEdit.text()
    
    def saveAsked(self):
        return self.__save_asked
    
    def acceptWithSave(self):
        self.__save_asked = True
        self.accept()
    
    
class SnapshotsDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_list_snapshots.Ui_Dialog()
        self.ui.setupUi(self)
        
        self._signaler.reply_auto_snapshot.connect(
            self.ui.checkBoxAutoSnapshot.setChecked)
        self._signaler.snapshots_found.connect(self.addSnapshots)
        
        self.snapshots = []
        self.main_snap_group = SnapGroup()
        
        self.ui.snapshotsList.setHeaderHidden(True)
        self.ui.snapshotsList.currentItemChanged.connect(
            self.currentItemChanged)
        
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
    
    def currentItemChanged(self, current, previous):
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(
           bool(current and current.data(0, Qt.UserRole))) 
    
    def decodeTimeString(self, time_str):
        while time_str.endswith('_'):
            time_str = time_str[:-1]
        
        label = ''
        strs = time_str.split('_')
        
        if len(strs) > 6:
            time_str = ''
            i = 0
            for stri in strs:
                if i < 6: 
                    time_str += "%s_" % stri
                else:
                    label += "%s_" % stri
                i+=1
                
            time_str = time_str[:-1]
            label = label[:-1]
            
        return (time_str, label)
    
    def addSnapshots(self, snaptexts):
        for snaptext in snaptexts:
            if not snaptext:
                continue
            
            time_str_full, line_change, rw_time_str_full_sess = \
                snaptext.partition('\n')
            rw_time_str_full, line_change, session_name = \
                rw_time_str_full_sess.partition('\n')
            
            time_str, two_points, label = time_str_full.partition(':')
            rw_time_str, two_points, rw_label = rw_time_str_full.partition(':')
            
            utc_date_time = QDateTime.fromString(time_str, 'yyyy_M_d_h_m_s')
            utc_rw_date_time = QDateTime.fromString(rw_time_str,
                                                'yyyy_M_d_h_m_s')
            utc_date_time.setTimeSpec(Qt.OffsetFromUTC)
            utc_rw_date_time.setTimeSpec(Qt.OffsetFromUTC)
            
            date_time = None
            rw_date_time = None
            
            if utc_date_time.isValid():
                date_time = utc_date_time.toLocalTime()
            
            if utc_rw_date_time.isValid():
                rw_date_time = utc_rw_date_time.toLocalTime()
            
            snapshot = Snapshot(date_time)
            snapshot.text = snaptext
            snapshot.label = label
            snapshot.rewind_date_time = rw_date_time
            snapshot.rewind_label = rw_label
            snapshot.session_name = session_name
            
            self.main_snap_group.add(snapshot)
            
        self.main_snap_group.sort()
        
        self.ui.snapshotsList.clear()
        
        for snapshot in self.main_snap_group.snapshots:
            item = snapshot.makeItem(GROUP_MAIN)
            self.ui.snapshotsList.addTopLevelItem(item)
            
        self.ui.buttonBox.button(QDialogButtonBox.Ok).setEnabled(False)
        self.ui.snapshotsList.clearSelection()
    
    def getSelectedSnapshot(self):
        item = self.ui.snapshotsList.currentItem()
        full_str = item.data(0, Qt.UserRole)
        snapshot_ref = full_str.partition('\n')[0].partition(':')[0]
        
        return snapshot_ref
    
    def showEvent(self, event):
        ChildDialog.showEvent(self, event)
        
        if RS.settings.value('hide_snapshots_info'):
            return
        
        info_dialog = SnapshotsInfoDialog(self)
        info_dialog.exec()
        
        if info_dialog.hasToBeHiddenNextTime():
            RS.settings.setValue('hide_snapshots_info', True)
        
    
class SnapshotsInfoDialog(ChildDialog):
    def __init__(self, parent):
        ChildDialog.__init__(self, parent)
        self.ui = ui_snapshots_info.Ui_Dialog()
        self.ui.setupUi(self)
    
    def hasToBeHiddenNextTime(self):
        return self.ui.checkBox.isChecked()

class SessionSnapshotsDialog(SnapshotsDialog):
    def __init__(self, parent):
        SnapshotsDialog.__init__(self, parent)
        
        self.ui.pushButtonSnapshotNow.clicked.connect(self.takeSnapshot)
        
        self.toDaemon('/ray/session/list_snapshots')
        
        self.ui.checkBoxAutoSnapshot.stateChanged.connect(
            self.setAutoSnapshot)
        
    def takeSnapshot(self):
        dialog = TakeSnapshotDialog(self)
        dialog.exec()
        if dialog.result():
            snapshot_label = dialog.getSnapshotName()
            with_save = dialog.saveAsked()
            self.toDaemon('/ray/session/take_snapshot', snapshot_label,
                          int(with_save))
    
    def setAutoSnapshot(self, bool_snapshot):
        self.toDaemon('/ray/session/set_auto_snapshot', int(bool_snapshot))

class ClientSnapshotsDialog(SnapshotsDialog):
    def __init__(self, parent, client):
        SnapshotsDialog.__init__(self, parent)
        self.ui.pushButtonSnapshotNow.hide()
        self.ui.checkBoxAutoSnapshot.hide()
        
        self.client = client
        
        self.toDaemon('/ray/client/list_snapshots', client.client_id)
        



#!/usr/bin/python3 -u

from PyQt5.QtCore import QCoreApplication, QObject, QTimer, pyqtSignal
from PyQt5.QtXml import QDomDocument
import sys, signal, os, time

#from shared import *
import jacklib
import nsm_client
import ray

connection_list = []
saved_connections = []
port_list = []

PORT_MODE_OUTPUT = 0
PORT_MODE_INPUT  = 1
PORT_MODE_NULL   = 2

PORT_TYPE_AUDIO = 0
PORT_TYPE_MIDI  = 1
PORT_TYPE_NULL  = 2

file_path = ""

is_dirty = False

pending_connection = False

def signalHandler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        app.quit()

class JackPort(object):
    #is_new is used to prevent reconnections when a disconnection has not been saved and one new port append. 
    slots = ['id', 'name', 'mode', 'type', 'is_new']

class ConnectTimer(QObject):
    def __init__(self):
        self.timer = QTimer()
        self.timer.setInterval(200)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(makeMayConnections)
    
    def start(self):
        self.timer.start()
        
def portExists(name, mode, port_type):
    for port in port_list:
        if port.name == name and port.mode == mode and port.type == port_type:
            return True
    return False

def setDirtyClean():
    global is_dirty
    is_dirty = False
    NSMServer.sendDirtyState(False)

def timerDirtyFinish():
    global is_dirty
    
    if is_dirty:
        return
    
    if isDirtyNow():
        is_dirty = True
        NSMServer.sendDirtyState(True)

def isDirtyNow():
    for connection in connection_list:
        if not connection in saved_connections:
            return True
        
    output_ports = []
    input_ports  = []
    
    for port in port_list:
        if port.mode == PORT_MODE_OUTPUT:
            output_ports.append(port.name)
        elif port.mode == PORT_MODE_INPUT:
            input_ports.append(port.name)
            
    for connection in saved_connections:
        if connection in connection_list:
            continue
        
        if connection[0] in output_ports and connection[1] in input_ports:
            return True
        
    return False

class DirtyChecker(QObject):
    timer = QTimer()
    
    def __init__(self):
        self.timer.setInterval(500)
        self.timer.setSingleShot(True)
        self.timer.timeout.connect(timerDirtyFinish)
            
    def start(self):
        self.timer.start()

def readyToConnect():
    pass

class Signaler(nsm_client.NSMSignaler):
    port_added   = pyqtSignal(str, int, int)
    port_removed = pyqtSignal(str, int, int)
    port_renamed = pyqtSignal(str, str, int, int)
    connection_added   = pyqtSignal(str, str)
    connection_removed = pyqtSignal(str, str)

def JackShutdownCallback(arg=None):
    app.quit()
    return 0
    
def JackPortRegistrationCallback(portId, registerYesNo, arg=None):
    portPtr = jacklib.port_by_id(jack_client, portId)
    portFlags = jacklib.port_flags(portPtr)
    port_name = str(jacklib.port_name(portPtr), encoding="utf-8")
    
    port_mode = PORT_MODE_NULL
    
    if portFlags & jacklib.JackPortIsInput:
        port_mode = PORT_MODE_INPUT
    elif portFlags & jacklib.JackPortIsOutput:
        port_mode = PORT_MODE_OUTPUT
    
    port_type = PORT_TYPE_NULL
    
    portTypeStr = str(jacklib.port_type(portPtr), encoding="utf-8")
    if portTypeStr == jacklib.JACK_DEFAULT_AUDIO_TYPE:
        port_type = PORT_TYPE_AUDIO
    elif portTypeStr == jacklib.JACK_DEFAULT_MIDI_TYPE:
        port_type = PORT_TYPE_MIDI
            
    if registerYesNo:
        signaler.port_added.emit(port_name, port_mode, port_type)
    else:
        signaler.port_removed.emit(port_name, port_mode, port_type)
        
    return 0

def JackPortRenameCallback(portId, oldName, newName, arg=None):
    portPtr = jacklib.port_by_id(jack_client, portId)
    portFlags = jacklib.port_flags(portPtr)
    
    port_mode = PORT_MODE_NULL
    
    if portFlags & jacklib.JackPortIsInput:
        port_mode = PORT_MODE_INPUT
    elif portFlags & jacklib.JackPortIsOutput:
        port_mode = PORT_MODE_OUTPUT
        
    port_type = PORT_TYPE_NULL
    
    portTypeStr = str(jacklib.port_type(portPtr), encoding="utf-8")
    if portTypeStr == jacklib.JACK_DEFAULT_AUDIO_TYPE:
        port_type = PORT_TYPE_AUDIO
    elif portTypeStr == jacklib.JACK_DEFAULT_MIDI_TYPE:
        port_type = PORT_TYPE_MIDI
        
    signaler.port_renamed.emit(str(oldName, encoding='utf-8'),
                               str(newName, encoding='utf-8'),
                               port_mode,
                               port_type)
    
    return 0
    

def JackPortConnectCallback(port_id_A, port_id_B, connect_yesno, arg=None):
    port_ptr_A = jacklib.port_by_id(jack_client, port_id_A)
    port_ptr_B = jacklib.port_by_id(jack_client, port_id_B)
     
    port_str_A = str(jacklib.port_name(port_ptr_A), encoding="utf-8")
    port_str_B = str(jacklib.port_name(port_ptr_B), encoding="utf-8")
    
    if connect_yesno:
        signaler.connection_added.emit(port_str_A, port_str_B)
    else:   
        signaler.connection_removed.emit(port_str_A, port_str_B)
        
    return 0

def portAdded(port_name, port_mode, port_type):
    port = JackPort()
    port.name   = port_name
    port.mode   = port_mode
    port.type   = port_type
    port.is_new = True
    
    port_list.append(port)
    
    connect_timer.start()
    
def portRemoved(port_name, port_mode, port_type):
    for i in range(len(port_list)):
        port = port_list[i]
        if (port.name == port_name
                and port.mode == port_mode
                and port.type == port_type):
            break
    else:
        return
    
    port_list.__delitem__(i)
    
def portRenamed(old_name, new_name, port_mode, port_type):
    for port in port_list:
        if (port.name == old_name
                and port.mode == port_mode
                and port.type == port_type):
            port.name   = new_name
            port.is_new = True
            connect_timer.start()
            break
   
def connectionAdded(port_str_A, port_str_B):
    connection_list.append((port_str_A, port_str_B))
    
    if pending_connection:
        makeMayConnections()
    
    if not (port_str_A, port_str_B) in saved_connections:
        dirty_checker.start()
    
def connectionRemoved(port_str_A, port_str_B):
    for i in range(len(connection_list)):
        if (connection_list[i][0] == port_str_A
                and connection_list[i][1] == port_str_B):
            connection_list.__delitem__(i)
            break
    
    dirty_checker.start()

def makeAllSavedConnections(port):
    if port.mode == PORT_MODE_OUTPUT:
        connectAllInputs(port)
    elif port.mode == PORT_MODE_INPUT:
        connectAllOutputs(port)

def connectAllInputs(port):
    if port.mode != PORT_MODE_OUTPUT:
        return
    
    input_ports  = []
    
    for jack_port in port_list:
        if jack_port.mode == PORT_MODE_INPUT:
            input_ports.append(jack_port.name)
            
    for connection in saved_connections:
        if connection in connection_list:
            continue
        
        if connection[0] == port.name and connection[1] in input_ports:
            jacklib.connect(jack_client, port.name, connection[1])

def connectAllOutputs(port):
    if port.mode != PORT_MODE_INPUT:
        return
    
    output_ports  = []
    
    for jack_port in port_list:
        if jack_port.mode == PORT_MODE_OUTPUT:
            output_ports.append(jack_port.name)
            
    for connection in saved_connections:
        if connection in connection_list:
            continue
        
        if connection[0] in output_ports and connection[1] == port.name:
            jacklib.connect(jack_client, connection[0], port.name)  

def makeMayConnections():
    output_ports = []
    input_ports  = []
    new_output_ports = []
    new_input_ports  = []
    
    for port in port_list:
        if port.mode == PORT_MODE_OUTPUT:
            output_ports.append(port.name)
            if port.is_new:
                new_output_ports.append(port.name)
                
        elif port.mode == PORT_MODE_INPUT:
            input_ports.append(port.name)
            if port.is_new:
                new_input_ports.append(port.name)
    
    global pending_connection
    one_connected = False
    
    for connection in saved_connections:
        if (not connection in connection_list
                and connection[0] in output_ports
                and connection[1] in input_ports
                and (connection[0] in new_output_ports
                     or connection[1] in new_input_ports)):
            
            if one_connected:
                pending_connection = True
                break
            
            jacklib.connect(jack_client, connection[0], connection[1])
            one_connected = True
    else:
        pending_connection = False
        
        for port in port_list:
            port.is_new = False
        
def c_char_p_p_to_list(c_char_p_p):
    i = 0
    retList = []

    if not c_char_p_p:
        return retList

    while True:
        new_char_p = c_char_p_p[i]
        if new_char_p:
            retList.append(str(new_char_p, encoding="utf-8"))
            i += 1
        else:
            break

    jacklib.free(c_char_p_p)
    return retList


def openFile(project_path, session_name, full_client_id):
    saved_connections.clear()
    
    global file_path
    file_path = "%s.xml" % project_path
    
    if os.path.isfile(file_path):
        try:
            file = open(file_path, 'r')
        except:
            print('unable to read file %s' % file_path, file=sys.stderr)
            app.quit()
            return
            
        xml = QDomDocument()
        xml.setContent(file.read())
        
        content = xml.documentElement()
        
        if content.tagName() != "RAY-JACKPATCH":
            file.close()
            NSMServer.openReply()
            return
        
        cte = content.toElement()
        node = cte.firstChild()
        
        while not node.isNull():
            el = node.toElement()
            if el.tagName() != "connection":
                continue
            
            port_from = el.attribute('from')
            port_to   = el.attribute('to')
            
            saved_connections.append((port_from, port_to))
            
            node = node.nextSibling()
        
        makeMayConnections()
        
    NSMServer.openReply()
    setDirtyClean()
    dirty_checker.start()
    

def saveFile():
    if not file_path:
        return
    
    for connection in connection_list:
        if not connection in saved_connections:
            saved_connections.append(connection)
    
    delete_list = []
    
    for i in range(len(saved_connections)):
        if (portExists(saved_connections[i][0],
                       PORT_MODE_OUTPUT, PORT_TYPE_AUDIO)
                and portExists(saved_connections[i][1],
                               PORT_MODE_INPUT, PORT_TYPE_AUDIO)):
            if not saved_connections[i] in connection_list:
                delete_list.append(i)
                
    delete_list.reverse()
    for i in delete_list:
        saved_connections.__delitem__(i)
    
    try:
        file = open(file_path, 'w')
    except:
        print('unable to write file %s' % file_path, file=sys.stderr)
        app.quit()
        return
        
    xml = QDomDocument()
    p = xml.createElement('RAY-JACKPATCH')
    
    for con in saved_connections:
        ct = xml.createElement('connection')
        ct.setAttribute('from', con[0])
        ct.setAttribute('to'  , con[1])
        p.appendChild(ct)
        
    xml.appendChild(p)
    
    file.write(xml.toString())
    file.close()
    
    NSMServer.saveReply()
    
    setDirtyClean()

if __name__ == '__main__':
    NSM_URL = os.getenv('NSM_URL')
    if not NSM_URL:
        print('Could not register as NSM client.', file=sys.stderr)
        sys.exit()
    
    daemon_address = ray.getLibloAddress(NSM_URL)
    
    jack_client = jacklib.client_open(
        "ray-patcher",
        jacklib.JackNoStartServer | jacklib.JackSessionID,
        None)
    
    if not jack_client:
        print('Unable to make a jack client !', file=sys.stderr)
        sys.exit()
        
        
    jacklib.set_port_registration_callback(jack_client,
                                           JackPortRegistrationCallback,
                                           None)
    jacklib.set_port_connect_callback(jack_client,
                                      JackPortConnectCallback,
                                      None)
    jacklib.set_port_rename_callback(jack_client,
                                     JackPortRenameCallback,
                                     None)
    jacklib.on_shutdown(jack_client, JackShutdownCallback, None)
    jacklib.activate(jack_client)
    
    signaler = Signaler()
    signaler.port_added.connect(portAdded)
    signaler.port_removed.connect(portRemoved)
    signaler.port_renamed.connect(portRenamed)
    signaler.connection_added.connect(connectionAdded)
    signaler.connection_removed.connect(connectionRemoved)
    signaler.server_sends_open.connect(openFile)
    signaler.server_sends_save.connect(saveFile)
    
    #makeMayConnections()
    
    NSMServer = nsm_client.NSMThread('ray-jackpatch', signaler,
                                     daemon_address, False)
    NSMServer.start()
    NSMServer.announce('JACK Connexions', ':dirty:switch:', 'ray-jackpatch')
    
    #connect signals
    signal.signal(signal.SIGINT , signalHandler)
    signal.signal(signal.SIGTERM, signalHandler)
    
    
    #get all currents Jack ports and connections
    portNameList = c_char_p_p_to_list(jacklib.get_ports(jack_client,
                                                        "", "", 0))
    
    for portName in portNameList:
        jack_port = JackPort()
        jack_port.name = portName
        
        portPtr = jacklib.port_by_name(jack_client, portName)
        portFlags = jacklib.port_flags(portPtr)
        
        if portFlags & jacklib.JackPortIsInput:
            jack_port.mode = PORT_MODE_INPUT
        elif portFlags & jacklib.JackPortIsOutput:
            jack_port.mode = PORT_MODE_OUTPUT
        else:
            jack_port.mode = PORT_MODE_NULL
        
        portTypeStr = str(jacklib.port_type(portPtr), encoding="utf-8")
        if portTypeStr == jacklib.JACK_DEFAULT_AUDIO_TYPE:
            jack_port.type = PORT_TYPE_AUDIO
        elif portTypeStr == jacklib.JACK_DEFAULT_MIDI_TYPE:
            jack_port.type = PORT_TYPE_MIDI
        else:
            jack_port.type = PORT_TYPE_NULL
        
        jack_port.is_new = True
        
        port_list.append(jack_port)
        
        if jacklib.port_flags(portPtr) & jacklib.JackPortIsInput:
            continue

        portConnectionNames = c_char_p_p_to_list(
                                jacklib.port_get_all_connections(jack_client,
                                                                 portPtr))

        for portConName in portConnectionNames:
            connection_list.append((portName, portConName))
            
    app = QCoreApplication(sys.argv)
    
    
    
    #needed for signals SIGINT, SIGTERM
    timer = QTimer()
    timer.start(200)
    timer.timeout.connect(lambda: None)
    
    connect_timer = ConnectTimer()
    dirty_checker = DirtyChecker()
    
    app.exec()
    
    jacklib.deactivate(jack_client)
    jacklib.client_close(jack_client)



#!/usr/bin/python3

import argparse
import liblo
import os
import signal
import sys
import time

from PyQt5.QtCore import QCoreApplication, QTimer, pyqtSignal, QObject

import ray

def signalHandler(sig, frame):
    if sig in (signal.SIGINT, signal.SIGTERM):
        QCoreApplication.quit()


class Signaler(QObject):
    done = pyqtSignal(int)

class OscServerThread(liblo.ServerThread):
    def __init__(self, daemon_port):
        liblo.ServerThread.__init__(self)
        self._daemon_port = daemon_port
        
    #@liblo.make_method('/reply', 'ss')
    #def replyMessage(self, path, args, types, src_addr):
        #reply_path, message = args
        #print('fko', reply_path, message)
        #sys.stdout.write("%s\n" % message)
        
        #if reply_path != '/nsm/server/list':
            #signaler.done.emit(0)
    
    #@liblo.make_method('/reply', 'sis')
    #def replyMessage2(self, path, args, types, src_addr):
        #reply_path, err, message = args
        #sys.stdout.write("%s\n" % message)
        
        #signaler.done.emit(- err)
    
    @liblo.make_method('/reply', None)
    def replyNone(self, path, args, types, src_addr):
        #print('fkorf', args)
        
        if len(args) == 2:
            reply_path, message = args
            #print('fko', reply_path, message)
            sys.stdout.write("%s\n" % message)
            
            if reply_path != '/nsm/server/list':
                signaler.done.emit(0)
        elif len(args) == 3:
            reply_path, err, message = args
            sys.stdout.write("%s\n" % message)
            
            signaler.done.emit(- err)
    
    @liblo.make_method('/error', 'sis')
    def errorMessage(self, path, args, types, src_addr):
        error_path, err, message = args
        sys.stdout.write('%s\n' % message)
        
        signaler.done.emit(- err)
    
    # this is very strange
    # nsmd sends a /nsm/server/list err Ok message when list is done.
    @liblo.make_method('/nsm/server/list', 'is')
    def nsmServerList(self, path, args, types, src_addr):
        signaler.done.emit(0)
    
    def toDaemon(self, *args):
        self.send(liblo.Address(self._daemon_port), *args)

def printHelp():
    sys.stdout.write(
        """control NSM server,
opions are
    add executable_name
        Adds a client to the current session.
    save
        Saves the current session.
    open project_name
        Saves the current session and loads a new session.
    new project_name
        Saves the current session and creates a new session.
    duplicate new_project
        Saves and closes the current session, makes a copy, and opens it.
    close
        Saves and closes the current session.
    abort
        Closes the current session *WITHOUT SAVING*  
    quit
        Saves and closes the current session and terminates the server.
    list 
        Lists available projects.
""")
    sys.exit(0)

#class finisher(QObject):
    #def finished(self, err_code):
        #global exit_code
        #exit_code = err_code
        #QCoreApplication.quit()
        
def finished(err_code):
    global exit_code
    exit_code = err_code
    QCoreApplication.quit()

if __name__ == '__main__':
    if len(sys.argv) <= 1:
        printHelp()
        sys.exit(100)
    
    operation = sys.argv[1]
    if not operation in ('add', 'save', 'open', 'new', 'duplicate', 
                         'close', 'abort', 'quit', 'list'):
        printHelp()
        sys.exit(100)
    
    arg_list = []
    if len(sys.argv) >= 3:
        arg_list = sys.argv[2:]
    
    if operation in ('add', 'open', 'new', 'duplicate'):
        if not arg_list:
            sys.stderr.write('missing argument after "%s"\n' % operation)
            sys.exit(100)
    
    nsm_port = 13555
    
    if ray.isOscPortFree(nsm_port):
        sys.stderr.write("No server at port %i\n" % nsm_port)
        sys.exit(100)
    
    exit_code = 0
    
    
    
    app = QCoreApplication(sys.argv)
    signaler = Signaler()
    signaler.done.connect(finished)
    
    osc_server = OscServerThread(nsm_port)
    osc_server.start()
    
    #connect SIGINT and SIGTERM
    signal.signal(signal.SIGINT,  signalHandler)
    signal.signal(signal.SIGTERM, signalHandler)
    
    #needed for SIGINT and SIGTERM
    timer = QTimer()
    timer.setInterval(200)
    timer.timeout.connect(lambda: None)
    timer.start()
    
    #time.sleep(0.201)
    osc_server.toDaemon('/nsm/server/%s' % operation, *arg_list)
    
    app.exec()
    osc_server.stop()
    
    sys.exit(exit_code)
    



#!/usr/bin/python3
import liblo
import os
import sys
from PyQt5.QtXml import QDomDocument

def noGoodArg():
    strerror = "%s currently only recognizes the following options:\n" \
                    % sys.argv[0]
    strerror += "save close abort\n"
    sys.stderr.write(strerror)
    sys.exit(1)

if __name__ == "__main__":
    if len(sys.argv) <= 1:
        noGoodArg()
        
    option = sys.argv[1]
    if not option in ('save', 'close', 'abort'):
        noGoodArg()
        
    multi_daemon_file = '/tmp/RaySession/multi-daemon.xml'
    if not os.path.isfile(multi_daemon_file):
        sys.exit(0)
        
    file = open(multi_daemon_file, 'r')
    xml = QDomDocument()
    xml.setContent(file.read())
    
    daemons_root = xml.documentElement()
    nodes = daemons_root.childNodes()
    for i in range(nodes.count()):
        node = nodes.at(i)
        node_el = node.toElement()
        port = node_el.attribute('port')
        liblo.send(liblo.Address(port), '/ray/session/%s' % option)



import locale
import os
import shutil
import socket
import subprocess
import sys
from PyQt5.QtCore import (QProcess, QProcessEnvironment, QTimer,
                          QObject, pyqtSignal, QDateTime)
from PyQt5.QtXml import QDomDocument
import ray
from daemon_tools import Terminal

def gitStringer(string):
    for char in (' ', '*', '?', '[', ']', '(', ')'):
        string = string.replace(char, "\\" + char)
      
    for char in ('#', '!'):
        if string.startswith(char):
            string = "\\" + string

    return string

def fullRefForGui(ref, name, rw_ref, rw_name='', ss_name=''):
    if ss_name:
        return "%s:%s\n%s:%s\n%s" % (ref, name, rw_ref, rw_name, ss_name) 
    return "%s:%s\n%s:%s" % (ref, name, rw_ref, rw_name) 

class Snapshoter(QObject):
    def __init__(self, session):
        QObject.__init__(self)
        self.session = session
        self.git_exec = 'git'
        self.gitdir = '.ray-snapshots'
        self.exclude_path = 'info/exclude'
        self.history_path = "session_history.xml"
        self.max_file_size = 50 #in Mb
        
        self.next_snapshot_name  = ''
        self.next_rw_snapshot = ''
        
        self.changes_checker = QProcess()
        self.changes_checker.readyReadStandardOutput.connect(
            self.changesCheckerStandardOutput)
        
        self.adder_process = QProcess()
        self.adder_process.finished.connect(self.save_step_1)
        self.adder_process.readyReadStandardOutput.connect(
            self.adderStandardOutput)
        
        self._adder_aborted = False
        
        self.git_process = QProcess()
        self.git_process.readyReadStandardOutput.connect(self.standardOutput)
        self.git_process.readyReadStandardError.connect(self.standardError)
        self.git_command = ''
        
        self._n_file_changed = 0
        self._n_file_treated = 0
        self._changes_counted = False
        
        self.next_function  = None
        self.error_function = None
    
    def changesCheckerStandardOutput(self):
        standard_output = self.changes_checker.readAllStandardOutput().data()
        self._n_file_changed += len(standard_output.decode().split('\n')) -1
    
    def adderStandardOutput(self):
        standard_output = self.adder_process.readAllStandardOutput().data()
        Terminal.snapshoterMessage(standard_output, ' add -A -v')
        
        if not self._n_file_changed:
            return
        
        self._n_file_treated += len(standard_output.decode().split('\n')) -1
        
        self.session.sendGui('/ray/gui/server/progress',
                             self._n_file_treated / self._n_file_changed)
    
    def standardError(self):
        standard_error = self.git_process.readAllStandardError().data()
        Terminal.snapshoterMessage(standard_error, self.git_command)
        
    def standardOutput(self):
        standard_output = self.git_process.readAllStandardOutput().data()
        Terminal.snapshoterMessage(standard_output, self.git_command)
    
    def getGitDir(self):
        if not self.session.path:
            raise NameError("attempting to save with no session path !!!")
        
        return "%s/%s" % (self.session.path, self.gitdir)
    
    def runGitProcess(self, *all_args):
        return self.runGitProcessAt(self.session.path, *all_args)
    
    def runGitProcessAt(self, spath, *all_args):
        self.git_command = ''
        for arg in all_args:
            self.git_command += ' %s' % arg
        
        err = ray.Err.OK
        
        git_args = self.getGitCommandListAt(spath, *all_args)
        self.git_process.start(self.git_exec, git_args)
        if not self.git_process.waitForFinished(2000):
            self.git_process.kill()
            err = ray.Err.SUBPROCESS_UNTERMINATED
        else:
            if self.git_process.exitStatus():
                err = ray.Err.SUBPROCESS_CRASH
            elif self.git_process.exitCode():
                err = ray.Err.SUBPROCESS_EXITCODE
        
        if err and self.error_function:
            self.error_function(err, ' '.join(all_args))
            
        return not(bool(err))
    
    def getGitCommandList(self, *args):
        return self.getGitCommandListAt(self.session.path, *args)
    
    def getGitCommandListAt(self, spath, *args):
        first_args = ['--work-tree', spath, '--git-dir',
                      "%s/%s" % (spath, self.gitdir)]
        
        return first_args + list(args)
    
    def getHistoryFullPath(self):
        return "%s/%s/%s" % (
                        self.session.path, self.gitdir, self.history_path)
    
    def getHistoryXmlDocumentElement(self):
        if not self.isInit():
            return None
        
        file_path = self.getHistoryFullPath()
        
        xml = QDomDocument()
            
        try:
            history_file = open(file_path, 'r')
            xml.setContent(history_file.read())
            history_file.close()
        except BaseException:
            return None
        
        SNS_xml = xml.documentElement()
        if SNS_xml.tagName() != 'SNAPSHOTS':
            return None
        
        return SNS_xml
        
    def list(self, client_id=""):
        SNS_xml = self.getHistoryXmlDocumentElement()
        if not SNS_xml:
            return []
        
        nodes = SNS_xml.childNodes()
        
        all_tags = []
        all_snaps = []
        prv_session_name = self.session.name
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            el = node.toElement()
            
            if client_id:
                client_nodes = node.childNodes()
                for j in range(client_nodes.count()):
                    client_node = client_nodes.at(j)
                    client_el = client_node.toElement()
                    if client_el.attribute('client_id') == client_id:
                        break
                else:
                    continue
            
            ref   = el.attribute('ref')
            name  = el.attribute('name')
            rw_sn = el.attribute('rewind_snapshot')
            rw_name = ""
            session_name = el.attribute('session_name')
            
            # don't list snapshot from client before session renamed
            if client_id and session_name != self.session.name:
                client = self.session.getClient(client_id)
                if (client
                    and client.prefix_mode == ray.PrefixMode.SESSION_NAME):
                        continue
            
            ss_name = ""
            if session_name != prv_session_name:
                ss_name = session_name
            
            prv_session_name = session_name
            
            if not ref.replace('_', '').isdigit():
                continue
                
            if '\n' in name:
                name = ""
            
            if not rw_sn.replace('_', '').isdigit():
                rw_sn = ""
            
            if rw_sn:
                for snap in all_snaps:
                    if snap[0] == rw_sn and not '\n' in snap[1]:
                        rw_name = snap[1]
                        break
            
            all_snaps.append((ref, name))
            all_tags.append(fullRefForGui(ref, name, rw_sn, rw_name, ss_name))
            
        return all_tags.__reversed__()
    
    def getTagDate(self):
        date_time = QDateTime.currentDateTimeUtc()
        date = date_time.date()
        time = date_time.time()
        
        tagdate = "%s_%s_%s_%s_%s_%s" % (
                    date.year(), date.month(), date.day(),
                    time.hour(), time.minute(), time.second())
        
        return tagdate
    
    def writeHistoryFile(self, date_str, snapshot_name='', rewind_snapshot=''):
        if not self.session.path:
            return ray.Err.NO_SESSION_OPEN
        
        file_path = self.getHistoryFullPath()
        
        xml = QDomDocument()
            
        try:
            history_file = open(file_path, 'r')
            xml.setContent(history_file.read())
            history_file.close()
        except:
            pass
        
        if xml.firstChild().isNull():
            SNS_xml = xml.createElement('SNAPSHOTS')
            xml.appendChild(SNS_xml)
        else:
            SNS_xml = xml.firstChild()
        
        snapshot_el = xml.createElement('Snapshot')
        snapshot_el.setAttribute('ref', date_str)
        snapshot_el.setAttribute('name', snapshot_name)
        snapshot_el.setAttribute('rewind_snapshot', rewind_snapshot)
        snapshot_el.setAttribute('session_name', self.session.name)
        snapshot_el.setAttribute('VERSION', ray.VERSION)
        
        for client in self.session.clients + self.session.removed_clients:
            client_el = xml.createElement('client')
            client.writeXmlProperties(client_el)
            client_el.setAttribute('client_id', client.client_id)
            
            for client_file_path in client.getProjectFiles():
                base_path = client_file_path.replace(
                    "%s/" % self.session.path, '', 1)
                file_xml = xml.createElement('file')
                file_xml.setAttribute('path', base_path)
                client_el.appendChild(file_xml)
            
            snapshot_el.appendChild(client_el)
            
        SNS_xml.appendChild(snapshot_el)
        
        try:
            history_file = open(file_path, 'w')
            history_file.write(xml.toString())
            history_file.close()
        except:
            return ray.Err.CREATE_FAILED
        
        return ray.Err.OK
    
    def getExcludeFileFullPath(self):
        return "%s/%s/%s" % (
                        self.session.path, self.gitdir, self.exclude_path)
    
    def writeExcludeFile(self):
        file_path = self.getExcludeFileFullPath()
        
        try:
            exclude_file = open(file_path, 'w')
        except:
            return ray.Err.CREATE_FAILED
        
        contents = ""
        contents += "# This file is generated by ray-daemon at each snapshot\n"
        contents += "# Don't edit this file.\n"
        contents += "# If you want to add/remove files managed by git\n"
        contents += "# Create/Edit .gitignore in the session folder\n"
        contents += "\n"
        contents += "%s\n" % self.gitdir
        contents += "\n"
        contents += "# Globally ignored extensions\n"
        
        session_ignored_extensions = ray.getGitIgnoredExtensions()
        session_ign_list = session_ignored_extensions.split(' ')
        session_ign_list = tuple(filter(bool, session_ign_list))
        
        # write global ignored extensions
        for extension in session_ign_list:
            contents+= "*%s\n" % extension 
            
            for client in self.session.clients:
                cext_list = client.ignored_extensions.split(' ')
                
                if not extension in cext_list:
                    contents += "!%s.%s/**/*%s\n" % (
                        gitStringer(client.getPrefixString()),
                        gitStringer(client.client_id),
                        extension)
                    contents += "!%s.%s.**/*%s\n" % (
                        gitStringer(client.getPrefixString()),
                        gitStringer(client.client_id),
                        extension)
                    
        contents += '\n'
        contents += "# Extensions ignored by clients\n"
        
        # write client specific ignored extension
        for client in self.session.clients:
            cext_list = client.ignored_extensions.split(' ')
            for extension in cext_list:
                if not extension:
                    continue
                
                if extension in session_ignored_extensions:
                    continue
                
                contents += "%s.%s/**/*%s\n" % (
                    gitStringer(client.getPrefixString()), 
                    gitStringer(client.client_id),
                    extension)
                
                contents += "%s.%s.**/*%s\n" % (
                    gitStringer(client.getPrefixString()), 
                    gitStringer(client.client_id),
                    extension)
        
        contents += '\n'
        contents += "# Too big Files\n"
        
        no_check_list = (self.gitdir)
        # check too big files
        for foldername, subfolders, filenames in os.walk(self.session.path):
            subfolders[:] = [d for d in subfolders if d not in no_check_list]
            
            if foldername == "%s/%s" % (self.session.path, self.gitdir):
                continue
            
            for filename in filenames:
                if filename.endswith(session_ign_list):
                    if os.path.islink(filename):
                        short_folder = foldername.replace(
                                        self.session.path + '/', '', 1)
                        line = gitStringer("%s/%s" % (short_folder, filename))
                        contents += '!%s\n' % line
                    # file with extension globally ignored but
                    # unignored by its client will not be ignored
                    # and that is well as this.
                    continue
                        
                if os.path.islink(filename):
                    continue
                
                try:
                    file_size = os.path.getsize(os.path.join(foldername,
                                                             filename))
                except:
                    continue
                
                if file_size > self.max_file_size*1024**2:
                    if foldername == self.session.path:
                        line = gitStringer(filename)
                    else:
                        short_folder = foldername.replace(
                                        self.session.path + '/', '', 1)
                        line = gitStringer("%s/%s" % (short_folder, filename))
                        
                    contents += "%s\n" % line
        
        try:
            exclude_file.write(contents)
            exclude_file.close()
        except:
            return ray.Err.CREATE_FAILED
        
        return ray.Err.OK
    
    def isInit(self):
        if not self.session.path:
            return False
        
        return os.path.isfile("%s/%s/%s" % (
                self.session.path, self.gitdir, self.exclude_path))
        
    def hasChanges(self):
        if not self.session.path:
            return False
        
        if not self.isInit():
            return True
        
        args = self.getGitCommandList('ls-files',
                                      '--exclude-standard',
                                      '--others',
                                      '--modified')
        if self.changes_checker.state():
            self.changes_checker.kill()
        
        self._n_file_changed = 0
        self._n_file_treated = 0
        self._changes_counted = True
        
        self.changes_checker.start(self.git_exec, args)
        self.changes_checker.waitForFinished(2000)
        
        return bool(self._n_file_changed)
    
    def canSave(self):
        if not self.session.path:
            return False
        
        if not self.isInit():
            if not self.runGitProcess('init'):
                return False
            
            user_name = os.getenv('USER')
            if not user_name:
                user_name = 'someone'
            
            machine_name = socket.gethostname()
            if not machine_name:
                machine_name = 'somewhere'
                
            if not self.runGitProcess('config', 'user.email', 
                                      '%s@%s' % (user_name, machine_name)):
                return False
            
            user_name = os.getenv('USER')
            if not user_name:
                user_name = 'someone'
            
            if not self.runGitProcess('config', 'user.name', user_name):
                return False
            
        if not self.isInit():
            return False
        
        return True
    
    def save(self, name='', rewind_snapshot='',
             next_function=None, error_function=None):
        self.next_snapshot_name  = name
        self._rw_snapshot = rewind_snapshot
        self.next_function = next_function
        self.error_function = error_function
        
        if not self.canSave():
            Terminal.message("can't snapshot")
            return
        
        err = self.writeExcludeFile()
        if err:
            if self.error_function:
                self.error_function(err)
            return 
            
        all_args = self.getGitCommandList('add', '-A', '-v')
        
        self._adder_aborted = False
        
        if not self._changes_counted:
            self.hasChanges()
            
        if self._n_file_changed:
            self.adder_process.start(self.git_exec, all_args)
        else:
            self.save_step_1()
            
        # self.adder_process.finished is connected to self.save_step_1
        
    def save_step_1(self):
        if self._adder_aborted:
            if self.next_function:
                self.next_function(aborted=True)
            return
        
        if self._n_file_changed:
            if not self.runGitProcess('commit', '-m', 'ray'):
                return
        
        self._changes_counted = False
        
        ref = self.getTagDate()
        
        if (self._n_file_changed
                or self.next_snapshot_name or self.next_rw_snapshot):
            if not self.runGitProcess('tag', '-a', ref, '-m', 'ray'):
                return 
        
        err = self.writeHistoryFile(ref, self.next_snapshot_name,
                                    self._rw_snapshot)
        if err:
            if self.error_function:
                self.error_function(err)
        
        # not really a reply, not strong.
        self.session.sendGui('/reply', '/ray/session/list_snapshots',
                             fullRefForGui(ref, self.next_snapshot_name,
                                           self._rw_snapshot))
        self.error_function = None
        
        if self.next_function:
            self.next_function()
        
    def load(self, spath, snapshot, error_function):
        self.error_function = error_function
        
        snapshot_ref = snapshot.partition('\n')[0].partition(':')[0]
        
        if not self.runGitProcessAt(spath, 'reset', '--hard'):
            return False
        
        if not self.runGitProcessAt(spath, 'checkout', snapshot_ref):
            return False
        return True
    
    def loadClientExclusive(self, client_id, snapshot, error_function):
        self.error_function = error_function
        
        SNS_xml = self.getHistoryXmlDocumentElement()
        if not SNS_xml:
            self.error_function(ray.Err.NO_SUCH_FILE,
                                self.getHistoryFullPath())
            return False
        
        nodes = SNS_xml.childNodes()
        
        client_path_list = []
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            el = node.toElement()
            
            if el.attribute('ref') != snapshot:
                continue
            
            client_nodes = node.childNodes()
            
            for j in range(client_nodes.count()):
                client_node = client_nodes.at(j)
                client_el = client_node.toElement()
                
                if client_el.attribute('client_id') != client_id:
                    continue
                
                file_nodes = client_node.childNodes()
                
                for k in range(file_nodes.count()):
                    file_node = file_nodes.at(k)
                    file_el = file_node.toElement()
                    file_path = file_el.attribute('path')
                    if file_path:
                        client_path_list.append(file_path)
        
        if not self.runGitProcess('reset', '--hard'):
            return False
        
        if not self.runGitProcess('checkout', snapshot, '--',
                                  *client_path_list):
            return False
        return True
        
    def abort(self):
        if not self.adder_process.state():
            return 
        
        self.setAutoSnapshot(False)
        
        self._adder_aborted = True
        self.adder_process.terminate()
    
    def setAutoSnapshot(self, bool_snapshot):
        auto_snap_file = "%s/%s/prevent_auto_snapshot" % (self.session.path,
                                                          self.gitdir)
        file_exists = bool(os.path.exists(auto_snap_file))
        
        if bool_snapshot:
            if file_exists:
                try:
                    os.remove(auto_snap_file)
                except PermissionError:
                    return
        else:
            if not file_exists:
                contents = "# This file prevent auto snapshots for this session (RaySession)\n"
                contents += "# remove it if you want auto snapshots back"
                
                try:
                    file = open(auto_snap_file, 'w')
                    file.write(contents)
                    file.close()
                except PermissionError:
                    return
                
    def isAutoSnapshotPrevented(self):
        auto_snap_file = "%s/%s/prevent_auto_snapshot" % (self.session.path, self.gitdir)
        
        return bool(os.path.exists(auto_snap_file))



import os
from PyQt5.QtXml import QDomDocument

instance = None

class MultiDaemonFile(object):
    def __init__(self, session, server):
        self.session = session
        self.server  = server
        
        self.file_path = '/tmp/RaySession/multi-daemon.xml'
        self.xml = QDomDocument()
    
        global instance
        instance = self
        
    @staticmethod
    def getInstance():
        return instance
    
    def pidExists(self, pid):
        if type(pid) == str:
            pid = int(pid)
        
        try:
            os.kill(pid, 0)
        except OSError:
            return False
        else:
            return True
    
    def removeFile(self):
        try:
            os.remove(self.file_path)
        except:
            return
    
    def openFile(self):
        if not os.path.exists(self.file_path):
            if not os.path.exists(os.path.dirname(self.file_path)):
                os.makedirs(os.path.dirname(self.file_path))
                
            return False
        
        try:
            file = open(self.file_path, 'r')
            self.xml.setContent(file.read())
            file.close()
            return True
            
        except:
            self.removeFile()
            return False
    
    def writeFile(self):
        try:
            file = open(self.file_path, 'w')
            file.write(self.xml.toString())
            file.close()
        except:
            return
    
    def setAttributes(self, element):
        element.setAttribute('net_daemon_id', self.server.net_daemon_id)
        element.setAttribute('root', self.session.root)
        element.setAttribute('session_path', self.session.path)
        element.setAttribute('pid', os.getpid())
        element.setAttribute('port', self.server.port)
        element.setAttribute('user', os.getenv('USER'))
    
    def update(self):
        if not self.openFile():
            ds = self.xml.createElement('Deamons')
            dm_xml = self.xml.createElement('Deamon')
            
            self.setAttributes(dm_xml)
            
            ds.appendChild(dm_xml)
            self.xml.appendChild(ds)
            
        else:
            found = False
            
            xml_content = self.xml.documentElement()
            
            nodes = xml_content.childNodes()
            for i in range(nodes.count()):
                node = nodes.at(i)
                dxe = node.toElement()
                pid  = dxe.attribute('pid')
                
                if pid.isdigit() and pid == str(os.getpid()):
                    self.setAttributes(dxe)
                    found = True
                
            if not found:
                dm_xml = self.xml.createElement('Deamon')
                self.setAttributes(dm_xml)
                self.xml.firstChild().appendChild(dm_xml)
                
        self.writeFile()
    
    def quit(self):
        if not self.openFile():
            return
        
        xml_content = self.xml.documentElement()
        nodes = xml_content.childNodes()
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            pid = dxe.attribute('pid')
            
            if pid.isdigit() and pid == str(os.getpid()):
                break
        else:
            return

        xml_content.removeChild(node)
        self.writeFile()
        
    def isFreeForRoot(self, daemon_id, root_path):
        if not self.openFile():
            return True
        
        xml_content = self.xml.documentElement()
        nodes = xml_content.childNodes()
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            if (dxe.attribute('net_daemon_id') == str(daemon_id) 
                    and dxe.attribute('root') == root_path):
                pid = dxe.attribute('pid')
                if pid.isdigit() and self.pidExists(int(pid)):
                    return False
            
        return True
    
    def isFreeForSession(self, session_path):
        if not self.openFile():
            return True
        
        xml_content = self.xml.documentElement()
        nodes = xml_content.childNodes()
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            if dxe.attribute('session_path') == session_path:
                pid = dxe.attribute('pid')
                if pid.isdigit() and self.pidExists(int(pid)):
                    return False
            
        return True
        
    def getAllSessionPaths(self):
        if not self.openFile():
            return []
        
        all_session_paths = []
        
        xml_content = self.xml.documentElement()
        nodes = xml_content.childNodes()
        
        for i in range(nodes.count()):
            node = nodes.at(i)
            dxe = node.toElement()
            spath = dxe.attribute('session_path')
            pid   = dxe.attribute('pid')
            if spath and pid.isdigit() and self.pidExists(int(pid)):
                all_session_paths.append(spath)
                
        return all_session_paths
